"""Test multiple source capabilities."""
import numpy as np
import numpy.testing as npt
import pyfar as pf
import pytest
import sparrowpy as sp

create_reference_files = False

@pytest.mark.parametrize('frequencies', [
    np.array([1000]),
    np.array([500,1000,2000]),
    ])
@pytest.mark.parametrize('receivers', [
    pf.Coordinates(
        [.25,.45,.75],
        [.45,.45,.45],
        [.75,.75,.75],
    ),
])
def test_multi_receiver(basicscene, frequencies,
                        receivers):
    """Check validity of multiple receiver output."""

    src = pf.Coordinates(.5, .5, .5)

    radi = run_basicscene(
        basicscene, src_pos=src, freqs=frequencies,
    )

    big_histo = radi.collect_energy_receiver_mono(
        receivers=receivers,
    )

    # assert correct dimensions of output histogram
    n_recs = receivers.cshape[0]

    assert big_histo.cshape[0] == n_recs
    assert big_histo.cshape[1] == frequencies.shape[0]

    # assert big histogram entries same as histograms for individual receiver
    for i in range(n_recs):
        small_histo = radi.collect_energy_receiver_mono(
            receivers=receivers[i],
            )

        npt.assert_array_almost_equal(
            big_histo[i].time, small_histo[0].time)


@pytest.mark.parametrize('src', [
    pf.Coordinates(2, 1.5, 1.5),
    ])
@pytest.mark.parametrize('rec', [
    pf.Coordinates(.5, 1, 2.5),
    pf.Coordinates(1, 2.5, 1.5),
    ])
@pytest.mark.parametrize('order', [
    10, 20,
    ])
@pytest.mark.parametrize('ps', [
    1, 1.5,
    ])
def test_reciprocity_shoebox(src,rec,order,ps):
    """Test if radiosity results are reciprocal in shoebox room."""
    X = 3
    Y = 3
    Z = 3
    patch_size = ps
    ir_length_s = .5
    sampling_rate = 200
    max_order_k = order
    speed_of_sound = 343
    etcs_new = []
    frequencies = np.array([1000])
    absorption = 0.
    walls = sp.testing.shoebox_room_stub(X, Y, Z)

    sc_src = pf.Coordinates(0, 0, 1)
    sc_rec = pf.Coordinates(0, 0, 1)

    for i in range(2):
        if i == 0:
            src_ = src
            rec_ = rec
        elif i == 1:
            src_ = rec
            rec_ = src


        ## initialize radiosity class
        radi = sp.DirectionalRadiosityFast.from_polygon(
            walls, patch_size)

        source_brdf = pf.Coordinates(0, 0, 1, weights=1)
        receivers_brdf = pf.Coordinates(0, 0, 1, weights=1)
        brdf = sp.brdf.create_from_scattering(
            source_brdf,
            receivers_brdf,
            pf.FrequencyData(
                np.ones((frequencies.size)), frequencies),
            pf.FrequencyData(
                np.zeros((frequencies.size))+absorption, frequencies),
            )

        # set directional scattering data
        radi.set_wall_brdf(
            np.arange(len(walls)), brdf, sc_src, sc_rec)

        # set air absorption
        radi.set_air_attenuation(
            pf.FrequencyData(
                np.zeros_like(frequencies),
                frequencies))


        # run simulation
        radi.bake_geometry()

        radi.init_source_energy(src_)

        radi.calculate_energy_exchange(
                            speed_of_sound=speed_of_sound,
                            etc_time_resolution=1/sampling_rate,
                            etc_duration=ir_length_s,
                            max_reflection_order=max_order_k,
                            )

        etc = radi.collect_energy_receiver_mono(rec_)

        # test energy at receiver
        etcs_new.append(etc.time[0])

    etcs_new = np.array(etcs_new)

    npt.assert_array_almost_equal(np.sum(etcs_new[1]), np.sum(etcs_new[0]))
    npt.assert_array_almost_equal(etcs_new[1], etcs_new[0])


@pytest.mark.parametrize('src', [
    [[2.,0,0], [-1, 0, 0], [0, 0, 1]],
    [[2.,2.,0], [-1, 0, 0], [0, 0, 1]],
    [[2.,0.,2.], [-1, 0, 0], [0, 0, 1]],
    ])
@pytest.mark.parametrize('rec', [
    [[1.,0.,0], [-1, 0, 0], [0, 0, 1]],
    [[2.,-2.,0], [-1, 0, 0], [0, 0, 1]],
    [[2.,0.,-2.], [-1, 0, 0], [0, 0, 1]],
    ])
def test_reciprocity_s2p_p2r(src,rec,method="universal"):
    """Check if radiosity implementation has source-receiver reciprocity."""
    wall = [sp.geometry.Polygon(
            [[0, -1, -1], [0, -1, 1],
            [0, 1, 1], [0, 1, -1]],
            [0, 0, 1], [1, 0, 0])]

    attenuation = pf.FrequencyData(
                np.zeros((1,1,1)),
                np.array([1000]))

    air_att = np.atleast_1d(attenuation.freq.squeeze())

    energy=[]

    for i in range(2):
        if i == 0:
            src_ = sp.geometry.SoundSource(src[0],src[1], src[2])
            rec_ = sp.geometry.Receiver(rec[0],rec[1], rec[2])
        elif i == 1:
            src_ = sp.geometry.SoundSource(rec[0],rec[1], rec[2])
            rec_ = sp.geometry.Receiver(src[0],src[1], src[2])

        if method == "universal":
            e_s,_ = sp.form_factor.universal._source2patch_energy_universal(
                                                    source_position=src_.position,
                                                    patches_center=np.array([wall[0].center]),
                                                    patches_points=np.array([wall[0].pts]),
                                                    source_visibility=np.ones((wall[0].center.shape[0]),dtype=bool),
                                                    air_attenuation=air_att,
                                                    n_bins=1,
                                                    )

            e_r = sp.form_factor.universal._patch2receiver_energy_universal(
                                                    receiver_pos=rec_.position,
                                                    patches_points=np.array([wall[0].pts]),
                                                    receiver_visibility=np.array([True]),
                                                    )


        elif method == "kang":
            e_s,_ = sp.form_factor.kang._source2patch_energy_kang(
                                                        source_position=src_.position,
                                                        patches_center=np.array([wall[0].center]),
                                                        patches_normal=np.array([wall[0].normal]),
                                                        air_attenuation=air_att,
                                                        patches_size=np.array([wall[0].size]),
                                                        n_bins=1,
                                                        )

            e_r = sp.form_factor.kang._patch2receiver_energy_kang(
                                                        patch_receiver_distance=np.array([(rec_.position-wall[0].center)]),
                                                        patches_normal=np.array([wall[0].normal]),
                                                        )

        e = e_s*e_r

        energy.append(e)

    npt.assert_array_almost_equal(energy[0], energy[1])


def run_basicscene(scene, src_pos, freqs):
    patch_size = scene["patch_size"]
    ir_length_s = scene["ir_length_s"]
    sampling_rate = scene["sampling_rate"]
    max_order_k = scene["max_order_k"]
    speed_of_sound = scene["speed_of_sound"]
    absorption = scene["absorption"]
    walls = scene["walls"]

    sc_src = pf.Coordinates(0, 0, 1, weights=1)
    sc_rec = pf.Coordinates(0, 0, 1, weights=1)

    ## initialize radiosity class
    radi = sp.DirectionalRadiosityFast.from_polygon(
        walls, patch_size)

    s = pf.FrequencyData(np.ones((freqs.size)), freqs)
    alpha = pf.FrequencyData(np.zeros((freqs.size))+absorption, freqs)
    brdf = sp.brdf.create_from_scattering(
        sc_src,
        sc_rec,
        s,
        alpha,
        )

    # set directional scattering data
    radi.set_wall_brdf(
        np.arange(len(walls)), brdf, sc_src, sc_rec)

    # set air absorption
    radi.set_air_attenuation(
        pf.FrequencyData(
            np.zeros_like(brdf.frequencies),
            brdf.frequencies))

    # run simulation
    radi.bake_geometry()

    radi.init_source_energy(src_pos)

    radi.calculate_energy_exchange(
            speed_of_sound=speed_of_sound,
            etc_time_resolution=1/sampling_rate,
            etc_duration=ir_length_s,
            max_reflection_order=max_order_k)

    return radi



"""Test radiosity module."""
import numpy as np
import numpy.testing as npt
import pytest
import pyfar as pf

import sparrowpy as sp



@pytest.mark.parametrize('walls', [
    # perpendicular walls
    [0, 2],
    # parallel walls
    [0, 1],
    ])
@pytest.mark.parametrize('patch_size', [
    0.5,
    ])
def test_form_factors_directivity_for_diffuse(
        sample_walls, walls, patch_size, sofa_data_diffuse):
    """Test form factor calculation for perpendicular walls."""
    wall_source = sample_walls[walls[0]]
    wall_receiver = sample_walls[walls[1]]
    walls = [wall_source, wall_receiver]

    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        walls, patch_size)
    data, sources, receivers = sofa_data_diffuse
    radiosity.set_wall_brdf(
        np.arange(len(walls)), data, sources, receivers)
    radiosity.set_air_attenuation(
        pf.FrequencyData(np.zeros_like(data.frequencies), data.frequencies))
    radiosity.bake_geometry()
    npt.assert_almost_equal(radiosity._form_factors_tilde.shape, (8, 8, 4, 4))
    # test _form_factors_tilde
    for i in range(radiosity._form_factors_tilde.shape[0]):
        for j in range(radiosity._form_factors_tilde.shape[0]):
            if i < j:
                npt.assert_almost_equal(
                    radiosity._form_factors_tilde[i, j, :, :],
                    radiosity._form_factors[i, j])
            else:
                npt.assert_almost_equal(
                    radiosity._form_factors_tilde[i, j, :, :],
                    radiosity._form_factors[j, i])


@pytest.mark.parametrize('patch_size', [1])
def test_order_vs_analytic(patch_size):
    """Test vs a perfect diffuse room, after Kuttruff."""
    # note that order k=0 means one reflection and k=1 means two reflections
    # (2nd order)
    X = 5
    Y = 3
    Z = 4
    length_histogram = .15
    time_resolution = 1/500
    max_order_k = 3
    speed_of_sound = 346.18
    receiver = pf.Coordinates(3, 1.5, 2)
    source = pf.Coordinates(2, 1.5, 2)

    absorption = 0.1

    sources = pf.Coordinates(0, 0, 1, weights=1)
    receivers = pf.Coordinates(0, 0, 1, weights=1)
    frequencies = np.array([500])
    brdf = sp.brdf.create_from_scattering(
        sources, receivers,
        pf.FrequencyData(1, frequencies),
        pf.FrequencyData(absorption, frequencies))
    walls = sp.testing.shoebox_room_stub(X, Y, Z)

    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        walls, patch_size)

    radiosity.set_wall_brdf(
        np.arange(len(walls)), brdf, sources, receivers)
    radiosity.set_air_attenuation(
        pf.FrequencyData(
            np.zeros_like(brdf.frequencies),
            brdf.frequencies))
    radiosity.bake_geometry()

    radiosity.init_source_energy(source)
    radiosity.calculate_energy_exchange(
        speed_of_sound=speed_of_sound,
        etc_time_resolution=time_resolution,
        etc_duration=length_histogram,
        max_reflection_order=max_order_k,
        recalculate=True)

    etc = radiosity.collect_energy_receiver_mono(receiver)

    # desired
    S = (2*X*Y) + (2*X*Z) + (2*Y*Z)
    A = S*absorption
    alpha_dash = A/S
    V = X*Y*Z
    E_reverb_analytical = 4 / A
    w_0 = E_reverb_analytical/ V # Kuttruff Eq 4.7
    t_0 = 0.03
    t = etc.times
    # Kuttruff Eq 4.10
    reverberation_analytic = w_0 * np.exp(+(
        speed_of_sound*S*np.log(1-alpha_dash)/(4*V))*(t-t_0))

    # compare histogram
    etc_radiosity_db = 10*np.log10(etc.time.flatten())
    etc_analytic_db = 10*np.log10(reverberation_analytic.flatten())

    # compare the diffuse part
    npt.assert_allclose(
        etc_radiosity_db[5:19],
        etc_analytic_db[5:19],
        atol=2) # maximal 2dB difference



"""Form factor models for Radiosity."""
from .universal import (
    patch2patch_ff_universal,
)
from .kang import (
    patch2patch_ff_kang,
)

__all__ = [
    "patch2patch_ff_universal",
    "patch2patch_ff_kang",
]



"""Methods and helpers for the form factor integration."""
try:
    import numba
    prange = numba.prange
except ImportError:
    numba = None
    prange = range
import numpy as np
import sparrowpy.geometry as geom


def load_stokes_entries(
    i_bpoints: np.ndarray, j_bpoints: np.ndarray) -> np.ndarray:
    """Load all the stokes form function values between two patches.

    Parameters
    ----------
    i_bpoints: np.ndarray
        list of points in patch i boundary (n_boundary_points_i , 3)

    j_bpoints: np.ndarray
        list of points in patch j boundary (n_boundary_points_j , 3)

    Returns
    -------
    form_mat: np.ndarray
        f function value matrix (n_boundary_points_i , n_boundary_points_j)

    """
    form_mat = np.zeros((len(i_bpoints) , len(j_bpoints)))

    for i in prange(i_bpoints.shape[0]):
        for j in prange(j_bpoints.shape[0]):
            form_mat[i][j] = np.log(np.linalg.norm(i_bpoints[i]-j_bpoints[j]))

    return form_mat

def stokes_integration(
    patch_i: np.ndarray, patch_j: np.ndarray, patch_i_area: float) -> float:
    """Calculate an estimation of the form factor between two patches.

    Computationally integrates a modified form function over
    the boundaries of both patches.
    The modified form function follows Stokes' theorem.

    The modified form function integral is calculated using a
    polynomial approximation based on sampled values.

    Parameters
    ----------
    patch_i : np.ndarray
        vertex coordinates of patch i (n_vertices, 3)

    patch_i_area: float
       area of patch i

    patch_j : np.ndarray
        vertex coordinates of patch j (n_vertices, 3)

    source_area: float
        area of the source patch

    approx_order: int
        polynomial order of the form function integration estimation

    Returns
    -------
    float
    form factor between two patches

    """
    i_bpoints, i_conn = _sample_boundary_regular(patch_i,
                                                         npoints=5)
    j_bpoints, j_conn = _sample_boundary_regular(patch_j,
                                                         npoints=5)

    subsecj = np.zeros((j_conn.shape[1]))
    subseci = np.zeros((i_conn.shape[1]))
    form_mat = np.zeros((i_bpoints.shape[0],j_bpoints.shape[0]))

    # first compute and store form function sample values
    form_mat = load_stokes_entries(i_bpoints, j_bpoints)

    # double polynomial integration (per dimension (x,y,z))
    outer_integral = 0
    inner_integral = np.zeros((len(i_bpoints),len(j_bpoints[0])))

    for dim in range(len(j_bpoints[0])): # for each dimension
        # integrate form function over each point on patch i boundary

        for i in range(len(i_bpoints)):   # for each point in patch i boundary
            for segj in j_conn:     # for each segment segj in patch j boundary

                xj = j_bpoints[segj][:,dim]

                if np.abs(xj[-1]-xj[0])>1e-3:
                    for k in range(len(segj)):
                        subsecj[k] = form_mat[i][segj[k]]

                    inner_integral[i][dim]+=_newton_cotes_4th(xj,subsecj)



        # integrate previously computed integral over patch i
        for segi in i_conn:   # for each segment segi in patch i boundary

            xi = i_bpoints[segi][:,dim]

            if np.abs(xi[-1]-xi[0])>1e-3:
                for k in range(len(segi)):
                    subseci[k] = inner_integral[segi[k]][dim]
                outer_integral+= _newton_cotes_4th(xi,subseci)

    return np.abs(outer_integral/(2*np.pi*patch_i_area))

def nusselt_analog(surf_origin, surf_normal,
                   patch_points, patch_normal) -> float:
    """Calculate the Nusselt analog for a single point.

    Projects a given receiver patch onto a hemisphere centered around a point
    on a source patch surface.
    The hemispherical projection is then projected onto the source patch plane.
    The area of this projection relative to the unit circle area is the
    differential form factor between the two patches.

    Parameters
    ----------
    surf_origin : np.ndarray
        point on source patch for differential form factor evaluation (3,)
        (global origin)

    surf_normal : np.ndarray
        normal of source patch (3,)

    patch_points : np.ndarray
        vertex coordinates of the receiver patch (n_vertices, 3)

    patch_normal: np.ndarray
        normal of receiver patch (3,)

    Returns
    -------
    Nusselt analog factor
    (differential form factor)

    """
    boundary_points, connectivity = _sample_boundary_regular(
                                                            patch_points,
                                                            npoints=3)

    hand = np.sign(np.dot(
                np.cross(patch_points[1]-patch_points[0],
                         patch_points[2]-patch_points[1]), patch_normal) )

    curved_area = 0

    sphPts = np.empty_like( boundary_points )
    projPts = np.empty_like( boundary_points )
    plnPts = np.empty( shape=(len(boundary_points),2) )

    for ii in prange(len(boundary_points)):
        # patch j points projected on the hemisphere
        sphPts[ii] = ( (boundary_points[ii]-surf_origin) /
                        np.linalg.norm(boundary_points[ii]-surf_origin) )

    rotmat = geom._rotation_matrix(n_in=surf_normal)

    for ii in prange(len(sphPts)):
        # points on the hemisphere projected onto patch plane
        plnPts[ii,:] = geom._matrix_vector_product(matrix=rotmat,
                                                      vector=sphPts[ii])[:-1]
        projPts[ii,:-1] = plnPts[ii,:]
        projPts[ii,-1] = 0.


    big_poly = geom._polygon_area(projPts[0::2])

    segmt=np.empty_like(connectivity[0])

    leftseg=np.empty((3,2))
    rightseg=np.empty((3,2))

    for jj in prange(connectivity.shape[0]):

        segmt = connectivity[jj]

        if (np.linalg.norm(np.cross(projPts[segmt[-1]],projPts[segmt[0]]))
                                                                    > 1e-6):

            # if the points on the segment span less than 90 degrees
            if np.dot( plnPts[segmt[-1]], plnPts[segmt[0]] ) >= 1e-6:
                curved_area += _area_under_curve(plnPts[segmt],order=2)

            # if points span over 90º, additional sampling is required
            else:
                mpoint = ( sphPts[segmt[0]] +
                          (sphPts[segmt[-1]] - sphPts[segmt[0]]) / 2 )

                # midpoint on the arc projected on the hemisphere
                marc = mpoint/np.linalg.norm(mpoint)
                a = sphPts[segmt[0]] + (marc - sphPts[segmt[0]]) / 2
                b = marc + (sphPts[segmt[-1]] - marc) / 2

                mpoint = geom._matrix_vector_product(matrix=rotmat,
                                                        vector=mpoint)[:-1]
                marc = geom._matrix_vector_product(matrix=rotmat,
                                                      vector=marc)[:-1]
                a = a/np.linalg.norm(a)
                a = geom._matrix_vector_product(matrix=rotmat,vector=a)[:-1]

                b = b/np.linalg.norm(b)
                b = geom._matrix_vector_product(matrix=rotmat,vector=b)[:-1]

                linArea = (np.linalg.norm(plnPts[segmt[-1]]-plnPts[segmt[0]])
                                              * np.linalg.norm(mpoint-marc)/2)

                leftseg[0] = plnPts[segmt[0]]
                leftseg[1] = a
                leftseg[2] = marc

                rightseg[0] = marc
                rightseg[1] = b
                rightseg[2] = plnPts[segmt[-1]]


                left =  _area_under_curve(leftseg, order=2)
                right = _area_under_curve(rightseg, order=2)
                curved_area += (linArea * np.sign(left) + left + right)

    return big_poly + hand*curved_area

def nusselt_integration(patch_i: np.ndarray, patch_j: np.ndarray,
                        patch_i_normal: np.ndarray, patch_j_normal: np.ndarray,
                        nsamples=2, random=False) -> float:
    """Estimate the form factor based on the Nusselt analogue.

    Integrates the differential form factor (Nusselt analogue output)
    over the surface of the source patch

    Parameters
    ----------
    patch_i: np.ndarray
        vertex coordinates of the source patch

    patch_j: np.ndarray
        vertex coordinates of the receiver patch

    patch_i_normal: np.ndarray
        source patch normal (3,)

    patch_j_normal: np.ndarray
        receiver patch normal (3,)

    patch_i_area: float
        source patch area

    patch_j_area: float
        receiver patch area

    nsamples: int
        number of receiver surface samples for integration

    random: bool
        determines the distribution of the samples on patch_i surface
        if True, the samples are randomly distributed in a uniform way
        if False, a regular sampling of the surface is performed

    Returns
    -------
    out: float
        form factor between patches i and j

    """
    if random:
        p0_array = _surf_sample_random(patch_i,nsamples)
    else:
        p0_array = _surf_sample_regulargrid(patch_i,nsamples)

    out = 0

    for i in prange(p0_array.shape[0]):
        out += nusselt_analog( surf_origin=p0_array[i],
                               surf_normal=patch_i_normal,
                               patch_points=patch_j,
                               patch_normal=patch_j_normal )

    out *= 1 / ( np.pi * len(p0_array) )

    return out


#/////////////////////////////////////////////////////////////////////////////////////#
#######################################################################################
### point-to-patch and patch-to-point
def pt_solution(point: np.ndarray, patch_points: np.ndarray, mode='source'):
    """Calculate the geometric factor between a point and a patch.

    applies a modified version of the Nusselt analogue,
    transformed for a -point- source rather than differential surface element.

    Parameters
    ----------
    point: np.ndarray
        source or receiver point

    patch_points: np.ndarray
        vertex coordinates of the patch

    mode: string
        determines if point is acting as a source ('source')
        or as a receiver ('receiver')

    Returns
    -------
    geometric factor

    """
    if mode == 'receiver':
        source_area = geom._polygon_area(patch_points)
    elif mode == 'source':
        source_area = 4

    npoints = len(patch_points)

    interior_angle_sum = 0

    patch_onsphere = np.zeros_like(patch_points)

    for i in range(npoints):
        patch_onsphere[i]= ( (patch_points[i]-point) /
                              np.linalg.norm(patch_points[i]-point) )

    for i in range(npoints):

        v0 = geom._sphere_tangent_vector(patch_onsphere[i],
                                              patch_onsphere[(i-1)%npoints])
        v1 = geom._sphere_tangent_vector(patch_onsphere[i],
                                              patch_onsphere[(i+1)%npoints])

        interior_angle_sum += np.arccos(np.dot(v0,v1))

    factor = interior_angle_sum - (len(patch_points)-2)*np.pi

    return factor / (np.pi*source_area)

###################################################
# integration
################# 1D , polynomial
def _poly_estimation_Lagrange(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Calculate Lagrange polynomial coefficients based on sample points.

    Computes coefficients of a polynomial curve passing through points (x,y)
    the order of the polynomial depends on the number of sample points
    input in the function. Uses the Lagrange method to estimate the polynomial.
        ex. a polynomial P estimated with 4 sample points:
            P4(x) = b[0]*x**3 + b[1]*x**2 + b[2]*x + b[3] = y

    Parameters
    ----------
    x: np.ndarray
        sample x-values
    y: np.ndarray
        sample y-values

    Returns
    -------
    b: np.ndarray
        polynomial coefficients

    """
    xmat = np.empty((len(x),len(x)))

    if np.abs(x[-1]-x[0])<1e-6:
        b = np.zeros(len(x))
    else:
        for i,xi in enumerate(x):
            for o in range(len(x)):
                xmat[i,len(x)-1-o] = xi**o

        b = geom._matrix_vector_product(np.linalg.inv(xmat), y)

    return b


def _poly_integration(c: np.ndarray, x: np.ndarray)-> float:
    """Integrate a polynomial curve.

    polynomial defined defined between x[0] and x[-1]
    with coefficients c
        ex. for a quadratic curve P2:
            P2(x) = c[0]*x**2 + c[1]*x + c[2]

    Parameters
    ----------
    c: np.ndarray
        polynomial coefficients
    x: np.ndarray
        sample points

    Returns
    -------
    out: float
        polynomial integral

    """
    out = 0

    for i in range(len(c)):
        out += c[i] * x[-1]**(len(c)-i) / (len(c)-i)
        out -= c[i] * x[0]**(len(c)-i) / (len(c)-i)

    return out

################# surface areas

def _area_under_curve(ps: np.ndarray, order=2) -> float:
    """Calculate the area under a polynomial curve.

    Curve sampled by a finite number of points ps on a common plane.

    Parameters
    ----------
    ps : np.ndarray
        sample points

    order : int
        polynomial order of the curve

    Returns
    -------
    area: float
        area under curve

    """
    # the order of the curve may be overwritten depending on the sample size
    order = min(order,len(ps)-1)

    # the vector between first and last sample (y==0) (new space's x axis)
    f  = ps[-1] - ps[0]

    rotation_matrix = np.array([[f[0],f[1]],[-f[1],f[0]]])/np.linalg.norm(f)

    x = np.zeros(order+1)
    y = np.zeros(order+1)

    for k in range(1,order+1):

        c = ps[k] - ps[0]  # translate point towards new origin

        # rotate point around origin to align with new axis
        cc = geom._matrix_vector_product(matrix=rotation_matrix,vector=c)

        x[k] = cc[0]
        y[k] = cc[1]


    coefs = _poly_estimation_Lagrange(x,y)
    area = _poly_integration(coefs,x) # area between curve and ps[-1] - ps[0]

    return area

def _newton_cotes_4th(x: np.ndarray,y):
    """Integrate 1D function after Boole's rule.

    Returns the approximate integral of a given function f(x)
    given 5 equally spaced samples (x,y).
    x samples the function input, while y samples the respective f(x) values.

    The integral is numerically estimated via the closed 4th-order
    Newton-Cotes formula (aka Boole's Rule).
    This formula *requires* 5 equidistant sample input.

    Parameters
    ----------
    x: np.ndarray(float)
        x-coordinate of the input samples.
    y: np.ndarray(float)
        f(x) of the input samples.
    """

    h=x[1]-x[0]
    return 2*h/45 *(7*y[0] +
                    32*y[1] +
                    12*y[2] +
                    32*y[3] +
                    7*y[4])

####################################################
# sampling
################# surface
def _surf_sample_random(el: np.ndarray, npoints=100):
    """Randomly sample points on the surface of a patch.

    ! currently only supports triangular, rectangular, or parallelogram patches

    Parameters
    ----------
    el : geometry.Polygon object
        patch to sample

    npoints : int
        number of sample points to generate

    Returns
    -------
    ptlist: np.ndarray
        list of sample points in patch el

    """

    ptlist=np.zeros((npoints,3))

    u = el[1]-el[0]
    v = el[-1]-el[0]

    for i in range(npoints):
        s = np.random.uniform()
        t = np.random.uniform()

        inside = s+t <= 1

        # if sample falls outside of triangular patch,
        # it is "reflected" back inside
        if len(el)==3 and not inside:
            s = 1-s
            t = 1-t

        ptlist[i] = s*u + t*v + el[0]

    return ptlist


def _surf_sample_regulargrid(el: np.ndarray, npoints=10):
    """Sample points on the surface of a patch using a regular distribution.

    over the directions defined by the patches' sides

    ! currently only supports triangular, rectangular, or parallelogram patches
    ! may not return exact number of requested points
                -- depends on the divisibility of the patch

    Parameters
    ----------
    el : geometry.Polygon object
        patch to sample

    npoints : int
        number of sample points to generate

    Returns
    -------
    out: np.ndarray
        list of sample points in patch el

    """
    u = el[1]-el[0]
    v = el[-1]-el[0]

    if len(el)==3:
        a = 2
    else:
        a = 1

    npointsx = int( round(np.linalg.norm(u) / np.linalg.norm(v) *
                                                np.sqrt(a*npoints)) )
    npointsz = int( round(np.linalg.norm(v) / np.linalg.norm(u) *
                                                np.sqrt(a*npoints)) )

    if npointsz==0:
        npointsz = 1
    if npointsx==0:
        npointsx = 1

    ptlist=[]

    tt = np.linspace(0,1-1/npointsx,npointsx)
    sstep =  1/(npointsx*2)
    tt += sstep

    tz = np.linspace(0,1-1/npointsz,npointsz)
    sstepz =  1/(npointsz*2)
    tz += sstepz

    thres = np.sqrt(sstepz**2+sstep**2)/2

    jj = 0

    for i,s in enumerate(tt):
        if len(el)==3:
            jj = i

        for t in tz[0:len(tz)-round(npointsz/npointsx*jj)]:

            inside = s+t <= 1-thres
            if not(len(el)==3 and not inside):
                ptlist.append(s*u + t*v + el[0])


    out = np.empty((len(ptlist), len(ptlist[0])))

    for i in prange(len(ptlist)):
        for j in prange(len(ptlist[0])):
            out[i][j] = ptlist[i][j]

    return out

################# boundary
def _sample_boundary_regular(el: np.ndarray, npoints=3):
    """Sample points on the boundary of a patch at fractional intervals.

    returns an array of points on the patch boundary (pts)
                                        and a connectivity array (conn)
    which stores a list of ordered indices of the points
    found on the same boundary segment.

    Parameters
    ----------
    el : geometry.Polygon object
        patch to sample

    npoints : int
        number of sample points per boundary segment (minimum 2)

    Returns
    -------
    pts: np.ndarray
        boundary sample points

    conn: np.ndarray(int)
        indices of pts corresponding to boundary segments
        (each row corresponds to the points in a single segment)

    """
    n_div = npoints - 1

    pts  = np.empty((len(el)*(npoints-1),len(el[0])))
    conn = np.empty((len(el),npoints), dtype=np.int8)

    for i in range(len(el)):

        conn[i][0]= (i*n_div)%(n_div*len(el))
        conn[i][-1]= (i*n_div+n_div)%(n_div*len(el))

        for ii in range(0,n_div):

            pts[i*n_div+ii,:]= (el[i] + ii*(el[(i+1)%len(el)]-el[i])/n_div)

            conn[i][ii]=(i*n_div+ii)%(n_div*len(el))


    return pts,conn.astype(np.int8)

if numba is not None:
    pt_solution = numba.njit(parallel=True)(pt_solution)
    nusselt_integration = numba.njit(parallel=False)(nusselt_integration)
    stokes_integration = numba.njit(parallel=False)(stokes_integration)
    nusselt_analog = numba.njit(parallel=False)(nusselt_analog)
    load_stokes_entries = numba.njit(parallel=True)(load_stokes_entries)
    _poly_estimation_Lagrange = numba.njit()(_poly_estimation_Lagrange)
    _poly_integration = numba.njit()(_poly_integration)
    _surf_sample_random = numba.njit()(_surf_sample_random)
    _surf_sample_regulargrid = numba.njit()(_surf_sample_regulargrid)
    _sample_boundary_regular = numba.njit()(_sample_boundary_regular)
    _area_under_curve = numba.njit()(_area_under_curve)
    _newton_cotes_4th = numba.njit()(_newton_cotes_4th)



"""Helper functions for handling of blender models."""
import sys
import os
from pathlib import Path
try:
    import bpy
    import bmesh
except ImportError:
    bpy = None
    bmesh = None
import numpy as np


class DotDict(dict):
    """dot.notation access to dictionary attributes."""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def read_geometry_file(blend_file: Path, angular_tolerance=3.):
    """Read blender file and return fine and rough mesh.

    Reads the input geometry from the blender file and reduces
    the mesh into a list of nodes (with spatial coordinates)
    and connectivity matrix (lists node indices belonging to same polygons).

    Furthermore, a rough mesh is generated in a similar fashion,
    by merging coplanar polygons into larger surface-wide polygons.

    Parameters
    ----------
    blend_file: Path
        path to blender file describing the
        scene geometry and setup

    angular_tolerance: float
        maximum angle in degree by which two patches are considered coplanar
        determines surfaces in simplified mesh

    Returns
    -------
    finemesh: dict
        includes vertex list and polygon connectivity matrix.

    roughmesh: dict
        mesh with reduced number of vertices.
        coplanar polygons in the input model are dissolved to
        create large surfaces.
        includes vertex list and polygon connectivity matrix.


    """
    if bpy is None:
        raise ImportError(
            "Blender is not installed. Please install "
            "Blender to use this function.")
    if os.path.splitext(blend_file)[-1] == ".blend":
        bpy.ops.wm.open_mainfile(filepath=str(blend_file))
    elif os.path.splitext(blend_file)[-1] ==".stl":
        bpy.ops.wm.stl_import(filepath=str(blend_file))
    else:
        NotImplementedError("Only .stl and .blend files are supported.")

    ensure_object_mode()

    objects = bpy.data.objects


    if "Geometry" not in objects:
        print("Geometry object not found in blend file")
        sys.exit()

    geometry = objects["Geometry"]


    # Creates file with only static geometric data of original blender file
    # without information about source and receiver
    bpy.ops.object.select_all(action="DESELECT")
    geometry.select_set(True)
    bpy.context.view_layer.objects.active = geometry


    # create bmesh from geometry
    out_mesh = bmesh.new()
    out_mesh.from_mesh(geometry.data)

    # dissolve coplanar faces for visibility check
    surfs = out_mesh.copy()
    bmesh.ops.dissolve_limit(surfs, angle_limit=angular_tolerance*np.pi/180,
                             verts=surfs.verts, edges=surfs.edges)

    finemesh = generate_connectivity(out_mesh)
    roughmesh = generate_connectivity(surfs)

    return finemesh, roughmesh


def ensure_object_mode():
    """Ensure Blender is in Object Mode."""
    if bpy is None:
        raise ImportError(
            "Blender is not installed. Please install "
            "Blender to use this function.")
    if bpy.context.object:
        if bpy.context.object.mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT')


def generate_connectivity(mesh: bmesh):
    """Generate node list and surf connectivity matrix.

    Parameters
    ----------
    mesh: bmesh
        mesh extracted from blender file

    Returns
    -------
    out_mesh: dict({
                    "conn":  list (n_polygons, :),
                    "verts": np.ndarray (n_nodes,3)
                    })
        mesh in reduced data representation.

    """
    if bmesh is None:
        raise ImportError(
            "Blender is not installed. Please install "
            "Blender to use this function.")
    out_mesh = {"conn":[], "verts": np.array([])}

    out_mesh["verts"] = np.array([v.co for v in mesh.verts])

    for f in mesh.faces:
        line = []
        for v in f.verts:
            line.append(v.index)
        out_mesh["conn"].append(line)

    return out_mesh



import numpy.testing as npt
import pytest
bpy = pytest.importorskip("bpy")
import sparrowpy.utils.blender as bh  # noqa: E402


@pytest.mark.parametrize("path",
                         ["./tests/test_data/cube.blend","./tests/test_data/cube.stl"])
def test_mesh_simplification(path):

    fine, rough = bh.read_geometry_file(path)

    npt.assert_array_equal(fine["verts"],rough["verts"])
    assert len(fine["conn"]) == len(rough["conn"])*2



"""Fixtures for the tests in the tests/ directory."""
import sparrowpy as sp
import pyfar as pf
import numpy as np
import pytest


@pytest.fixture
def brdf_s_0(tmp_path_factory):
    """Temporary small SOFA file.

    To be used when data needs to be read from a SOFA file for testing.
    Contains custom data for "Data_IR", "GLOBAL_RoomType" and
    "Data_SamplingRate_Units".

    Returns
    -------
    filename : SOFA file
        Filename of temporary SOFA file

    """
    filename = tmp_path_factory.mktemp("data") / "brdf_tmp.sofa"
    coords = pf.samplings.sph_gaussian(sh_order=1)
    coords = coords[coords.z > 0]
    sp.brdf.create_from_scattering(
        coords, coords, pf.FrequencyData(0, [100]), file_path=filename)
    return filename


@pytest.fixture
def brdf_s_1(tmp_path_factory):
    """Temporary small SOFA file.

    To be used when data needs to be read from a SOFA file for testing.
    Contains custom data for "Data_IR", "GLOBAL_RoomType" and
    "Data_SamplingRate_Units".

    Returns
    -------
    filename : SOFA file
        Filename of temporary SOFA file

    """
    filename = tmp_path_factory.mktemp("data") / "brdf_tmp.sofa"
    coords = pf.samplings.sph_gaussian(sh_order=1)
    coords = coords[coords.z > 0]
    sp.brdf.create_from_scattering(
        coords, coords, pf.FrequencyData(1, [100]), file_path=filename)
    return filename


@pytest.fixture
def sample_walls():
    """Return a list of 6 walls, which form a cube."""
    return sp.testing.shoebox_room_stub(1, 1, 1)


@pytest.fixture
def sofa_data_diffuse():
    """Return a diffuse brdf set for five octave bands."""
    gaussian = pf.samplings.sph_gaussian(sh_order=1)
    gaussian = gaussian[gaussian.z>0]
    sources = gaussian.copy()
    receivers = gaussian.copy()
    frequencies = np.array([125, 250, 500, 1000])
    brdf = sp.brdf.create_from_scattering(
        sources, receivers,
        pf.FrequencyData(np.ones_like(frequencies), frequencies))
    return (brdf, sources, receivers)


@pytest.fixture
def sofa_data_diffuse_full_third_octave():
    """Return a diffuse brdf set for the full third octave band."""
    gaussian = pf.samplings.sph_gaussian(sh_order=1)
    gaussian = gaussian[gaussian.z>0]
    sources = gaussian.copy()
    receivers = gaussian.copy()
    frequencies = np.array(
        [20.0, 25.0, 31.5, 40.0, 50.0, 63.0, 80.0, 100.0, 125.0, 160.0, 200.0,
         250.0, 315.0, 400.0, 500.0, 630.0, 800.0, 1000.0, 1250.0, 1600.0,
         2000.0, 2500.0, 3150.0, 4000.0, 5000.0, 6300.0, 8000.0, 10000.0,
         12500.0, 16000.0, 20000.0])
    brdf = sp.brdf.create_from_scattering(
        sources, receivers,
        pf.FrequencyData(np.ones_like(frequencies), frequencies))
    return (brdf, sources, receivers)


@pytest.fixture
def basicscene():
    scene = {}
    scene["patch_size"] = .2
    scene["ir_length_s"] = 1.
    scene["sampling_rate"] = 100
    scene["max_order_k"] = 10
    scene["speed_of_sound"] = 343
    scene["absorption"] = 0.1
    scene["walls"] = sp.testing.shoebox_room_stub(1, 1, 1)

    return scene



"""Test the universal form factor module."""
import pytest
import sparrowpy.geometry as geo
import numpy as np
import sparrowpy.testing.exact_ff_solutions as exact_solutions
from sparrowpy.sound_object import SoundSource, Receiver
from sparrowpy import PatchesKang
import sparrowpy.form_factor as form_factor
import sparrowpy as sp


@pytest.mark.parametrize("width", [1.])
@pytest.mark.parametrize("height", [1.,2.,3.,4])
@pytest.mark.parametrize("distance", [1.,2,3,4])
def test_parallel_facing_patches(width, height, distance):
    """Test universal form factor for equal facing parallel patches."""
    exact = exact_solutions.parallel_patches(width, height, distance)

    patch_1 = geo.Polygon(
        points=[[0, 0, 0], [width, 0, 0], [width, 0, height], [0, 0, height]],
        normal=[0, 1, 0],
        up_vector=[1, 0, 0],
    )

    patch_2 = geo.Polygon(
        points=[
            [0, distance, 0],
            [0, distance, height],
            [width, distance, height],
            [width, distance, 0],
        ],
        normal=[0, -1, 0],
        up_vector=[1, 0, 0],
    )

    univ = form_factor.patch2patch_ff_universal(
        patches_points=np.array([patch_1.pts,
                                 patch_2.pts]),
                          patches_normals=np.array(
                              [patch_1.normal, patch_2.normal]),
                          patches_areas=np.array([patch_1.area, patch_2.area]),
                          visible_patches=np.array([[0,1]]))

    rel = 100 * abs(univ[0,1] - exact) / exact

    assert  rel < 1.5


@pytest.mark.parametrize("width", [1.0, 2.0, 3.0])
@pytest.mark.parametrize("height", [1.0, 2.0, 3.0])
@pytest.mark.parametrize("length", [1.0])
def test_perpendicular_coincidentline_patches(width, height, length):
    """Test universal form factor for perpendicular patches sharing a side."""
    patch_1 = geo.Polygon(
        points=[
            [0, 0, 0],
            [0, length, 0],
            [0, length, height],
            [0, 0, height],
        ],
        normal=[1, 0, 0],
        up_vector=[1, 0, 0],
    )

    patch_2 = geo.Polygon(
        points=[[0, 0, 0],
                [width, 0, 0],
                [width, length, 0],
                [0, length, 0]],
        normal=[0, 0, 1],
        up_vector=[1, 0, 0],
    )

    exact = exact_solutions.perpendicular_patch_coincidentline(
        width, height, length,
    )

    univ = form_factor.patch2patch_ff_universal(
        patches_points=np.array([patch_1.pts,
                                 patch_2.pts]),
                          patches_normals=np.array(
                              [patch_1.normal, patch_2.normal]),
                          patches_areas=np.array([patch_1.area, patch_2.area]),
                          visible_patches=np.array([[0,1]]))

    rel = 100 * abs(univ[0,1] - exact) / exact

    assert rel < 2


@pytest.mark.parametrize("width1", [1.0, 2.0, 3.0])
@pytest.mark.parametrize("width2", [1.0, 2.0, 3.0])
@pytest.mark.parametrize("length1", [1.0, 2.0, 3.0])
@pytest.mark.parametrize("length2", [1.0, 2.0, 3.0])
def test_perpendicular_coincidentpoint_patches(
    width1, length1, width2, length2,
):
    """Test form factor for perpendicular patches w/ common vertex."""
    patch_1 = geo.Polygon(
        points=[
            [0., 0., 0.],
            [width2, 0., 0.],
            [width2, length2, 0.],
            [0., length2, 0.],
        ],
        normal=[0, 0, 1],
        up_vector=[0, 1, 0],
    )

    patch_2 = geo.Polygon(
        points=[
            [0., 0., 0.],
            [0., 0., width1],
            [0., -length1, width1],
            [0., -length1, 0.],
        ],
        normal=[1, 0, 0],
        up_vector=[1, 0, 0],
    )

    exact = exact_solutions.perpendicular_patch_coincidentpoint(
        width1, length2, width2, length1,
    )

    univ = form_factor.patch2patch_ff_universal(
        patches_points=np.array([patch_1.pts,
                                 patch_2.pts]),
                          patches_normals=np.array(
                              [patch_1.normal, patch_2.normal]),
                          patches_areas=np.array([patch_1.area, patch_2.area]),
                          visible_patches=np.array([[0,1]]))

    rel = 100 * abs(univ[0,1] - exact) / exact

    assert rel < 5

@pytest.mark.parametrize("X", [2.0, 3.0])
@pytest.mark.parametrize("Y", [1.0, 2.0])
@pytest.mark.parametrize("Z", [2.0])
def test_ff_energy_conservation(
    X,Y,Z,
):
    """Test form factor for perpendicular patches w/ common vertex."""

    walls = sp.testing.shoebox_room_stub(X, Y, Z)

    radi = sp.DirectionalRadiosityFast.from_polygon(walls, patch_size=1.)

    radi.bake_geometry()


    err = radi.n_patches-np.sum(radi._form_factors_tilde)
    assert err/radi.n_patches < 1e-2

    for i in range(radi.n_patches):
        err = 1-np.sum(radi._form_factors_tilde[i])
        assert err < 1e-2


@pytest.mark.parametrize("width", [1.0, 2.0, 3.0, 4.])
def test_different_areas(
    width,
):
    """Test reciprocity of form factors with different areas."""

    patch1 = np.array([[0.,0.,0.],
                       [width,0.,0.],
                       [width,1.,0.],
                       [0.,1.,0.]])

    patch2 = np.array([[0.,0.,1.],
                       [0.,1.,1.],
                       [width,1.,1.],
                       [width,0.,1.]])

    ff = form_factor.patch2patch_ff_universal(
        patches_points=np.array([patch1,patch2,patch2,patch1]),
        patches_normals=np.array(
            [[0.,0.,1.],[0.,0.,-1.],[0.,0.,-1.],[0.,0.,1.]]),
        patches_areas=np.array([width,1.,1.,width]),
        visible_patches=np.array([[0,1],[2,3]]))

    ff_tilde = sp.classes.RadiosityFast._form_factors_with_directivity_dim(
        visibility_matrix=np.array([[False,True,False,False],
                                [False,False,False,False],
                                [False,False,False,True],
                                [False,False,False,False]]),
        form_factors=ff,
        n_bins=1,
        patches_center=np.array([[width/2, .5, 0.],
                                    [.5,.5,1.],
                                    [.5,.5,1.],
                                    [width/2, .5, 0.]]),
        patches_area=np.array([width,1.,1.,width]),
        air_attenuation=None,
        patch_to_wall_ids=[0,1,1,0],
        scattering=None,
        scattering_index=None,
        sources=None,
        receivers=None,
    )

    assert ff_tilde[0,1,0,0]==ff_tilde[1,0,0,0]/width
    assert ff_tilde[2,3,0,0]/width==ff_tilde[3,2,0,0]
    assert np.abs(ff_tilde[0,1,0,0]-ff_tilde[3,2,0,0])<1e-9
    assert np.abs(ff_tilde[2,3,0,0]-ff_tilde[1,0,0,0])<1e-9


@pytest.mark.parametrize("side", [0.1, 0.2, 0.5, 1, 2])
@pytest.mark.parametrize(
    "source",
    [
        SoundSource(
            position=np.array([1, 7, 1]),
            view=np.array([0, -1, 0]),
            up=np.array([0, 0, 1]),
        ),
    ],
)
@pytest.mark.parametrize(
    "receiver",
    [
        Receiver(
            position=np.array([1, 1, 1]),
            view=np.array([0, -1, 0]),
            up=np.array([0, 0, 1]),
        ),
    ],
)
@pytest.mark.parametrize("patchsize", [0.1])
def test_point_surface_interactions(side, source, receiver, patchsize):
    """Test source-to-patch and patch-to-receiver factor calculation."""
    sr = 1000
    c = 343

    absor_factor = 0.1

    patch_pos = geo.Polygon(
        points=[[0, 0, 0], [side, 0, 0], [side, 0, side], [0, 0, side]],
        normal=[0, 1, 0],
        up_vector=[1, 0, 0],
    )

    patch = PatchesKang(
        polygon=patch_pos,
        max_size=patchsize * side,
        other_wall_ids=[],
        wall_id=[0],
        absorption=absor_factor,
    )

    patch.init_energy_exchange(
        0, 0.1, source, sampling_rate=sr, speed_of_sound=c,
    )

    patch = source_cast(src=source, rpatch=patch, absor=absor_factor)

    receiver_cast(receiver, patch, sr, c)


def source_cast(src, rpatch, absor):
    """Cast and test source-to-patch factor calculation."""
    nuss = form_factor.integration.pt_solution(point=src.position,
                                       patch_points=rpatch.pts)

    true = sum(rpatch.E_matrix[rpatch.E_matrix != 0])

    rel_error_nuss = abs(true - nuss * (1 - absor)) / true * 100

    assert rel_error_nuss < 1.0

    return rpatch


def receiver_cast(rcv, patch, sr, c):
    """Cast and test patch-to-receiver factor calculation."""
    true_rec_energy = np.sum(
        patch.energy_at_receiver(
            receiver=rcv,
            max_order=0,
            speed_of_sound=c,
            sampling_rate=sr,
        ),
    )

    patch_energy = np.sum(patch.E_matrix)

    nuss = form_factor.integration.pt_solution(
            point=rcv.position, patch_points=patch.pts, mode="receiver",
            ) * patch_energy

    rel_error_nuss = abs(true_rec_energy - nuss) / true_rec_energy * 100

    assert rel_error_nuss < 1.0



"""Module for the radiosity simulation."""
import matplotlib as mpl
import numpy as np
import pyfar as pf
import sofar as sf
from sparrowpy.geometry import Polygon, SoundSource


class PatchesKang(Polygon):
    """Class representing patches of a polygon."""

    patches: list[Polygon]
    other_wall_ids: list[int]  # ids of the patches
    form_factors: np.ndarray
    E_matrix: np.ndarray
    wall_id: int
    scattering: np.ndarray
    absorption: np.ndarray
    n_bins: int
    max_size: float
    E_sampling_rate: int
    E_n_samples: int
    sound_attenuation_factor: np.ndarray

    def __init__(
            self, polygon, max_size, other_wall_ids, wall_id,
            scattering=1, absorption=0.1, sound_attenuation_factor=0,
            E_matrix=None):
        """Init Directional Patches.

        Parameters
        ----------
        polygon : Polygon
            Wall that should be distributed into patches.
        max_size : float
            maximal patchsize in meters.
        other_wall_ids : list[int]
            Ids of the other walls.
        wall_id : int
            id of this Patch
        scattering : np.ndarray, optional
            Scattering coefficient of this wall for each freqeuncy
            bin, by default 1
        absorption : np.ndarray, optional
            Absorption coefficient of this wall for each freqeuncy
            bin, by default 0.1
        sound_attenuation_factor : _type_, optional
            Air attenuation factor for each frequncy bin, by default None
        E_matrix : np.ndarray, optional
            Energy exchange results, if already calcualted , by default None

        """
        Polygon.__init__(self, polygon.pts, polygon.up_vector, polygon.normal)
        min_point = np.min(polygon.pts, axis=0)
        max_point = np.max(polygon.pts, axis=0)
        size = max_point - min_point
        patch_nums = np.array([int(n) for n in size/max_size])
        real_size = size/patch_nums
        self.scattering = np.atleast_1d(scattering)
        self.absorption = np.atleast_1d(absorption)
        self.sound_attenuation_factor = np.atleast_1d(sound_attenuation_factor)
        assert self.scattering.size == self.absorption.size
        assert self.scattering.size == self.sound_attenuation_factor.size
        self.n_bins = self.absorption.size

        if patch_nums[2] == 0:
            x_idx = 0
            y_idx = 1
        if patch_nums[1] == 0:
            x_idx = 0
            y_idx = 2
        if patch_nums[0] == 0:
            x_idx = 1
            y_idx = 2

        x_min = np.min(polygon.pts.T[x_idx])
        y_min = np.min(polygon.pts.T[y_idx])

        patches = []
        for i_x in range(patch_nums[x_idx]):
            for i_y in range(patch_nums[y_idx]):
                points = polygon.pts.copy()
                points[0, x_idx] = x_min + i_x * real_size[x_idx]
                points[0, y_idx] = y_min + i_y * real_size[y_idx]
                points[1, x_idx] = x_min + (i_x+1) * real_size[x_idx]
                points[1, y_idx] = y_min + i_y * real_size[y_idx]
                points[3, x_idx] = x_min + i_x * real_size[x_idx]
                points[3, y_idx] = y_min + (i_y+1) * real_size[y_idx]
                points[2, x_idx] = x_min + (i_x+1) * real_size[x_idx]
                points[2, y_idx] = y_min + (i_y+1) * real_size[y_idx]

                patch = Polygon(points, polygon.up_vector, polygon.normal)
                patches.append(patch)

        self.patches = patches
        self.other_wall_ids = np.atleast_1d(np.array(
            other_wall_ids, dtype=int))

        self.wall_id = wall_id

        self.max_size = max_size

        if E_matrix is not None:
            self.E_matrix = np.array(E_matrix)
            self.E_n_samples = self.E_matrix.shape[3]

    def to_dict(self) -> dict:
        """Convert this object to dictionary. Used for read write."""
        return {
            'pts': self.pts,
            'up_vector': self.up_vector,
            'normal': self.normal,
            'max_size': self.max_size,
            'other_wall_ids': self.other_wall_ids,
            'wall_id': self.wall_id,
            'scattering': self.scattering,
            'absorption': self.absorption,
            'sound_attenuation_factor': self.sound_attenuation_factor,
            'E_matrix': self.E_matrix,
        }

    @classmethod
    def from_dict(cls, input_dict: dict):
        """Create an object from a dictionary. Used for read write."""
        return cls(
            Polygon.from_dict(input_dict),
            input_dict['max_size'],
            input_dict['other_wall_ids'],
            input_dict['wall_id'],
            input_dict['scattering'],
            input_dict['absorption'],
            input_dict['sound_attenuation_factor'],
            E_matrix=input_dict['E_matrix'])

    def plot(self, ax: mpl.axes.Axes = None, color=None):
        """Plot the patches."""
        for patch in self.patches:
            patch.plot_point(ax, color)
        self.plot_view_up(ax, color)

    def init_energy_exchange(
            self, max_order_k, ir_length_s, source, sampling_rate,
            speed_of_sound):
        """Initialize the energy exchange Matrix with source energy.

        It init the matrix self.E_matrix and add source energy after (6).

        Parameters
        ----------
        max_order_k : int
            max order of energy exchange iterations.
        ir_length_s : float
            length of the impulse response in seconds.
        source : SoundSource
            sound source with ``sound_power`` and ``position``
        sampling_rate : int, optional
            Sampling rate of impulse response.
        speed_of_sound : float, optional
            speed of sound in m/s.

        """
        self.E_sampling_rate = sampling_rate
        self.E_n_samples = int(ir_length_s*sampling_rate)

        self.E_matrix = np.zeros((
            self.n_bins,
            max_order_k+1,  # order of energy exchange
            len(self.patches),  # receiver ids of patches
            self.E_n_samples,  # impulse response G_k(t)_receiverpatch
            ))

        for i_receiver, receiver_patch in enumerate(self.patches):
            source_pos = source.position.copy()
            receiver_pos = receiver_patch.center.copy()

            distance = np.linalg.norm(receiver_pos-source_pos)
            delay_seconds = distance/speed_of_sound
            delay_samples = int(delay_seconds*self.E_sampling_rate)

            if np.abs(receiver_patch.normal[2]) > 0.99:
                i = 2
                indexes = [0, 1, 2]
            elif np.abs(receiver_patch.normal[1]) > 0.99:
                indexes = [2, 0, 1]
                i = 1
            elif np.abs(receiver_patch.normal[0]) > 0.99:
                i = 0
                indexes = [1, 2, 0]
            else:
                raise AssertionError()
            offset = receiver_pos[i]
            source_pos[i] = np.abs(source_pos[i] - offset)
            receiver_pos[i] = np.abs(receiver_pos[i] - offset)
            dl = receiver_pos[indexes[0]]
            dm = receiver_pos[indexes[1]]
            dn = receiver_pos[indexes[2]]
            dd_l = receiver_patch.size[indexes[0]]
            dd_m = receiver_patch.size[indexes[1]]
            S_x = source_pos[indexes[0]]
            S_y = source_pos[indexes[1]]
            S_z = source_pos[indexes[2]]
            energy = _init_energy_exchange(
                dl, dm, dn, dd_l, dd_m, S_x, S_y, S_z,
                source.sound_power, self.absorption, distance,
                self.sound_attenuation_factor, self.n_bins)
            self.E_matrix[:, 0, i_receiver, delay_samples] += energy


    def calculate_energy_exchange(
            self, patches_list, current_order_k, speed_of_sound,
            E_sampling_rate):
        """Calculate the energy exchange for a given order.

        It implements formula 18 and save the result in self.E_matrix.

        Parameters
        ----------
        patches_list : list[Patches]
            list of all patches
        current_order_k : int
            Order k
        speed_of_sound : int, optional
            speed of sound in m/s.
        E_sampling_rate : int, optional
            Sampling rate of histogram.

        """
        k = current_order_k # real k 1 .. max_k
        for i_receiver, receiver_patch in enumerate(self.patches):
            receiver = receiver_patch.center

            for wall_id in self.other_wall_ids:
                wall = patches_list[wall_id]
                for i_source, source_patch in enumerate(wall.patches):
                    source = source_patch.center

                    # distance between source and receiver patches
                    distance = np.linalg.norm(receiver-source)
                    delay_seconds = distance/speed_of_sound

                    # sample delay for the given distance
                    delay_samples = int(delay_seconds*E_sampling_rate)

                    for i_frequency in range(self.n_bins):
                        # access energy matrix of source patch, previous order
                        A_k_minus_1 = wall.E_matrix[
                            i_frequency, k-1, i_source, :]

                        # delay IR by delay_samples
                        A_k_minus1_delay = _add_delay(
                            A_k_minus_1, delay_samples)

                        # find form factor of source and receiver patches
                        form_factor = wall.get_form_factor(
                            patches_list, i_source, self.wall_id, i_receiver)
                        alpha = self.absorption[i_frequency]

                        # multiply delayed IR by form factor
                        energy = A_k_minus1_delay * form_factor \
                            * self.scattering[i_frequency] * \
                                (1-alpha) * np.exp(
                                    -self.sound_attenuation_factor[
                                        i_frequency] * distance)

                        # add energy to energy matrix of self
                        self.E_matrix[i_frequency, k, i_receiver, :] += energy
                        #

    def calculate_form_factor(self, patches_list) -> None:
        """Calculate the form factors between patches.

        Parameters
        ----------
        patches_list : list of patches
            List of patches.
        M : float, optional
            Air attenuation factor in Np/m, by default 0
        alpha : float, optional
            absorption coefficient of wall, by default 0.1

        """
        num_other_patches = np.sum([
            len(patches_list[i].patches) for i in self.other_wall_ids])
        self.form_factors = np.zeros((len(self.patches), num_other_patches))
        for i_source, source_patch in enumerate(self.patches):
            i_receiver_offset = 0
            for receiver_wall_id in self.other_wall_ids:
                receiver_wall = patches_list[receiver_wall_id]
                for i_receiver, receiver_patch in enumerate(
                        receiver_wall.patches):

                    difference = np.abs(
                        receiver_wall.center-self.center)
                    # calculation of form factors
                    dot_product = np.dot(
                        receiver_patch.normal, source_patch.normal)

                    dd_l = self.max_size
                    dd_m = self.max_size
                    dd_n = self.max_size

                    if dot_product == 0:  # orthogonal
                        source_center = source_patch.center
                        receiver_center = receiver_patch.center

                        if np.abs(source_patch.normal[0]) > 1e-5:
                            idx_source = {2, 1}
                            dl = source_center[2]
                            dm = source_center[1]
                        elif np.abs(source_patch.normal[1]) > 1e-5:
                            idx_source = {2, 0}
                            dl = source_center[2]
                            dm = source_center[0]
                        elif np.abs(source_patch.normal[2]) > 1e-5:
                            idx_source = {0, 1}
                            dl = source_center[1]
                            dm = source_center[0]

                        if np.abs(receiver_patch.normal[0]) > 1e-5:
                            idx_receiver = {2, 1}
                            dl_prime = receiver_center[1]
                            dn_prime = receiver_center[2]
                        elif np.abs(receiver_patch.normal[1]) > 1e-5:
                            idx_receiver = {0, 2}
                            dl_prime = receiver_center[0]
                            dn_prime = receiver_center[2]
                        elif np.abs(receiver_patch.normal[2]) > 1e-5:
                            idx_receiver = {0, 1}
                            dl_prime = receiver_center[1]
                            dn_prime = receiver_center[0]
                        idx_l = tuple(idx_receiver.intersection(idx_source))
                        idx_s = tuple(idx_source.difference(set(idx_l)))
                        idx_r = tuple(idx_receiver.difference(set(idx_l)))
                        dm = np.abs(
                            source_center[idx_s]-receiver_center[idx_s])
                        dl = source_center[idx_l]
                        dl_prime = receiver_center[idx_l]
                        dn_prime = np.abs(
                            source_center[idx_r]-receiver_center[idx_r])

                        d = np.sqrt( ( (dl - dl_prime) ** 2 ) + ( dm ** 2 ) + (
                            dn_prime ** 2) )

                        # Equation 13
                        A = ( dm - ( 0.5 * dd_m ) ) / ( np.sqrt( ( (
                            dl - dl_prime) ** 2 ) + ( ( dm - (
                                0.5 * dd_m ) ) ** 2 ) + ( dn_prime ** 2) ) )

                        # Equation 14
                        B_num = dm + (0.5 * dd_m)
                        B_denum =  np.sqrt(( np.square(dl - dl_prime) ) + (
                            np.square(dm + (0.5*dd_m)) ) + (
                                np.square(dn_prime)) )
                        B = B_num/B_denum

                        # Equation 15
                        one = np.arctan( np.abs( ( dl - (
                            0.5*dd_l) - dl_prime ) / (dn_prime) ) )
                        two = np.arctan( np.abs( ( dl + (
                            0.5*dd_l) - dl_prime ) / (dn_prime) ) )

                        k = -1 if np.abs(dl - dl_prime) < 1e-12 else 1

                        theta = np.abs( one - (k*two) )

                        # Equation 11
                        ff =  (
                            1 / (2 * np.pi) ) * (np.abs(
                                (A ** 2) - (B ** 2) )) * theta

                    else:
                        # parallel
                        if difference[0] > 1e-5:
                            dl = receiver_patch.center[1]
                            dm = receiver_patch.center[0]
                            dn = receiver_patch.center[2]
                            dl_prime = source_patch.center[1]
                            dm_prime = source_patch.center[0]
                            dn_prime = source_patch.center[2]
                        elif difference[1] > 1e-5:
                            dl = receiver_patch.center[0]
                            dm = receiver_patch.center[1]
                            dn = receiver_patch.center[2]
                            dl_prime = source_patch.center[0]
                            dm_prime = source_patch.center[1]
                            dn_prime = source_patch.center[2]
                        elif difference[2] > 1e-5:
                            dl = receiver_patch.center[1]
                            dm = receiver_patch.center[2]
                            dn = receiver_patch.center[0]
                            dl_prime = source_patch.center[1]
                            dm_prime = source_patch.center[2]
                            dn_prime = source_patch.center[0]
                        else:
                            raise AssertionError()

                        d = np.sqrt(
                            ( (dl - dl_prime) ** 2 ) +
                            ( (dn - dn_prime) ** 2 ) +
                            ( (dm - dm_prime) ** 2 ) )
                        # Equation 16
                        ff =  ( dd_l * dd_n * ( (
                            dm-dm_prime) ** 2 ) ) / ( np.pi * ( d**4 ) )

                    index_rec = i_receiver + i_receiver_offset
                    self.form_factors[i_source, index_rec] = ff

                i_receiver_offset += len(receiver_wall.patches)

    def get_form_factor(
            self, patches_list, source_path_id, receiver_wall_id,
            receiver_patch_id):
        """Return form factor.

        Parameters
        ----------
        patches_list : list of Patches
            patches list
        source_path_id : int
            _description_
        receiver_wall_id : int
            receiver wall id
        receiver_patch_id : _type_
            _description_

        Returns
        -------
        _type_
            _description_

        """
        i_receiver_offset = 0
        for other_wall in self.other_wall_ids:
            if other_wall == receiver_wall_id:
                i_receiver_ff = receiver_patch_id
                break
            wall = patches_list[other_wall]
            i_receiver_offset += len(wall.patches)

        return self.form_factors[
            source_path_id, i_receiver_ff + i_receiver_offset]

    def energy_at_receiver(
            self, max_order, receiver,
            speed_of_sound, sampling_rate):
        """Calculate the energy at the receiver.

        this is supposed to be from just one wall

        Parameters
        ----------
        max_order : _type_
            _description_
        sound_source : _type_
            _description_
        receiver : _type_
            _description_
        speed_of_sound : float, optional
            _description_
        sampling_rate : int, optional
            _description_

        Returns
        -------
        _type_
            _description_

        """
        energy_response = np.zeros((self.n_bins, self.E_n_samples))

        for i_source, source_patch in enumerate(self.patches):
            source_pos = source_patch.center
            receiver_pos = receiver.position

            difference = np.abs(receiver_pos-source_pos)
            R = np.linalg.norm(source_pos-receiver_pos)
            delay = int(R/speed_of_sound * sampling_rate)

            cos_xi = np.abs(np.sum(source_patch.normal*difference)) / \
                np.linalg.norm(source_pos-receiver_pos)

            for k in range(max_order+1):
                for i_frequency in range(self.n_bins):
                    energy = self.E_matrix[i_frequency, k, i_source, :]
                    delayed_energy = _add_delay(energy, delay)

                    # Equation 20
                    factor = cos_xi * (np.exp(
                        -self.sound_attenuation_factor[i_frequency]*R)) / (
                            np.pi * R**2)
                    receiver_energy = delayed_energy * factor
                    if factor <0:
                        print(factor)

                    energy_response[i_frequency, ...] += receiver_energy

        return energy_response


class PatchesDirectionalKang(PatchesKang):
    """Class representing patches with directional scattering behaviour."""

    directivity_data: np.ndarray
    directivity_sources: pf.Coordinates
    directivity_receivers: pf.Coordinates

    def __init__(
            self, polygon, max_size, other_wall_ids, wall_id,
            data, sources, receivers, absorption=None,
            sound_attenuation_factor=None, already_converted=False,
            E_matrix=None):
        """Init Directional Patches.

        Parameters
        ----------
        polygon : Polygon
            Wall that should be distributed into patches.
        max_size : float
            maximal patchsize in meters.
        other_wall_ids : list[int]
            Ids of the other walls.
        wall_id : int
            id of this Patch
        data : pf.FrequencyData
            Directional Data
        sources : pf.Coordinates
            source positions of the directivity data
        receivers : pf.Coordinates
            receiver positions of the directivity data
        absorption : np.ndarray, optional
            Absorption coefficient of this wall for each freqeuncy
            bin, by default None
        sound_attenuation_factor : _type_, optional
            Air attenuation factor for each frequncy bin, by default None
        already_converted : bool, optional
            is ``sources`` and ``receivers`` already converted, by default
            False.
        E_matrix : np.ndarray, optional
            Energy exchange results, if already calcualted , by default None

        """
        self.directivity_data = data
        if absorption is None:
            absorption = np.zeros_like(data.frequencies)+0.1
        if sound_attenuation_factor is None:
            sound_attenuation_factor = np.zeros_like(data.frequencies)
        PatchesKang.__init__(
            self, polygon, max_size, other_wall_ids, wall_id,
            scattering=np.ones_like(data.frequencies),
            absorption=absorption,
            sound_attenuation_factor=sound_attenuation_factor,
            E_matrix=E_matrix,
        )
        self.directivity_data.freq = np.abs(self.directivity_data.freq)
        self.directivity_sources = sources
        self.directivity_receivers = receivers
        if not already_converted:
            self.directivity_data.freq *= np.pi
            o1 = pf.Orientations.from_view_up(
                polygon.normal, polygon.up_vector)
            o2 = pf.Orientations.from_view_up([0, 0, 1], [1, 0, 0])
            o_diff = o1.inv()*o2
            euler = o_diff.as_euler('xyz', True).flatten()
            self.directivity_receivers.rotate('xyz', euler)
            self.directivity_receivers.radius = 1
            self.directivity_sources.rotate('xyz', euler)
            self.directivity_sources.radius = 1

    @classmethod
    def from_sofa(cls, polygon, max_size, other_wall_ids, wall_id,
            wall_directivity_path, absorption=None,
            sound_attenuation_factor=None):
        """Create object with directional data from sofa."""
        sofa = sf.read_sofa(wall_directivity_path, True, False)
        data, sources, receivers = pf.io.convert_sofa(sofa)
        sources.weights = sofa.SourceWeights
        receivers.weights = sofa.ReceiverWeights
        assert sources.weights is not None, "No source weights in sofa file"
        assert receivers.weights is not None, \
            "No receiver weights in sofa file"
        return cls(
            polygon, max_size, other_wall_ids, wall_id,
            data, sources, receivers, absorption=absorption,
            sound_attenuation_factor=sound_attenuation_factor)

    def to_dict(self) -> dict:
        """Convert this object to dictionary. Used for read write."""
        return {
            **PatchesKang.to_dict(self),
            'directivity_data': self.directivity_data.freq,
            'directivity_data_frequencies': self.directivity_data.frequencies,
            'directivity_sources': self.directivity_sources.cartesian,
            'directivity_receivers': self.directivity_receivers.cartesian,
            'directivity_sources_weights': self.directivity_sources.weights,
            'directivity_receivers_weights': \
                self.directivity_receivers.weights,
        }

    @classmethod
    def from_dict(cls, input_dict: dict):
        """Create an object from a dictionary. Used for read write."""
        data = pf.FrequencyData(
            input_dict['directivity_data'],
            input_dict['directivity_data_frequencies'])
        sources = pf.Coordinates(
            np.array(input_dict['directivity_sources']).T[0],
            np.array(input_dict['directivity_sources']).T[1],
            np.array(input_dict['directivity_sources']).T[2],
            weights=input_dict['directivity_sources_weights'])
        receivers = pf.Coordinates(
            np.array(input_dict['directivity_receivers']).T[0],
            np.array(input_dict['directivity_receivers']).T[1],
            np.array(input_dict['directivity_receivers']).T[2],
            weights=input_dict['directivity_receivers_weights'])
        return cls(
            Polygon.from_dict(input_dict),
            max_size=input_dict['max_size'],
            other_wall_ids=input_dict['other_wall_ids'],
            wall_id=input_dict['wall_id'],
            data=data,
            sources=sources,
            receivers=receivers,
            absorption=input_dict['absorption'],
            sound_attenuation_factor=input_dict['sound_attenuation_factor'],
            E_matrix=input_dict['E_matrix'],
            already_converted=True,
            )

    def init_energy_exchange(
            self, max_order_k, ir_length_s, source,
            sampling_rate, speed_of_sound):
        """Initialize the energy exchange Matrix with source energy.

        Parameters
        ----------
        max_order_k : int
            max order of energy exchange iterations.
        ir_length_s : float
            length of the impulse response in seconds.
        source : SoundSource
            Sound source with ``sound_power`` and ``position``
        sampling_rate : int
            Sample rate of histogram, by default 1000 -> 1ms
        speed_of_sound : float,
            speed of sound in m/s.

        """
        PatchesKang.init_energy_exchange(
            self, max_order_k, ir_length_s, source, sampling_rate,
            speed_of_sound=speed_of_sound)
        test = self.E_matrix.copy()
        self.E_matrix = self.E_matrix[..., np.newaxis]
        self.E_matrix = np.tile(
            self.E_matrix, self.directivity_receivers.csize)
        assert np.sum(test-self.E_matrix[..., 0]) == 0
        patches_center = np.array([patch.center for patch in self.patches])
        difference = source.position - patches_center
        difference = pf.Coordinates(
            difference.T[0], difference.T[1], difference.T[2])
        source_idx = self.directivity_sources.find_nearest(difference)[0][0]
        # get directional scattering coefficient for
        # all incident angles x receiver
        for i_frequency in range(self.directivity_data.n_bins):
            scattering = self.directivity_data.freq[source_idx, :, i_frequency]
            if len(self.patches) == 1:
                self.E_matrix[i_frequency, 0, 0, :, :] *= np.abs(
                    scattering)
            else:
                for i_patch in range(len(self.patches)):
                    self.E_matrix[i_frequency, 0, i_patch, :, :] *= np.abs(
                        scattering[i_patch, :])

    def calculate_energy_exchange(
            self, patches_list, current_order_k, speed_of_sound,
            E_sampling_rate):
        """Calculate the energy exchange for a given order.

        It implements formula 18 and save the result in self.E_matrix.

        Parameters
        ----------
        patches_list : list[Patches]
            list of all patches
        current_order_k : int
            Order k
        speed_of_sound : int, optional
            speed of sound in m/s.
        E_sampling_rate : int, optional
            Sampling rate of histogram, e.g. 1000 Hz -> 1ms.

        """
        k = current_order_k # real k 1 .. max_k
        for i_receiver, receiver_patch in enumerate(self.patches):
            receiver = receiver_patch.center

            for wall_id in self.other_wall_ids:
                wall = patches_list[wall_id]
                for i_source, source_patch in enumerate(wall.patches):
                    source = source_patch.center

                    # distance between source and receiver patches
                    distance = np.linalg.norm(receiver-source)
                    delay_seconds = distance/speed_of_sound

                    # sample delay for the given distance
                    delay_samples = int(delay_seconds*E_sampling_rate)

                    # find form factor of source and receiver patches
                    form_factor = wall.get_form_factor(
                        patches_list, i_source, self.wall_id, i_receiver)

                    for i_frequency in range(self.directivity_data.n_bins):
                        # access energy matrix of source patch, previous order
                        A_k_minus_1 = wall.E_matrix[
                            i_frequency, k-1, i_source, :, :]

                        # delay IR by delay_samples
                        A_k_minus1_delay = _add_delay(
                            A_k_minus_1, delay_samples, 0)

                        # multiply delayed IR by form factor
                        alpha = self.absorption[i_frequency]

                        energy = A_k_minus1_delay * form_factor * (
                            1-alpha) * np.exp(
                                -self.sound_attenuation_factor[i_frequency] \
                                    * distance)

                        # find directional scattering coefficient
                        difference = source_patch.center - \
                            receiver_patch.center
                        difference = pf.Coordinates(
                            difference.T[0], difference.T[1], difference.T[2])
                        source_idx = self.directivity_sources.find_nearest(
                            difference)[0][0]
                        # get directional scattering coefficient for
                        # all incident angles x receiver
                        scattering = self.directivity_data.freq[
                            source_idx, :, i_frequency]

                        energy *= scattering

                        # add energy to energy matrix of self
                        self.E_matrix[
                            i_frequency, k, i_receiver, :, :] += energy

    def energy_at_receiver(
            self, max_order, receiver,
            speed_of_sound, sampling_rate):
        """Calculate the energy at the receiver.

        this is supposed to be from just one wall

        Parameters
        ----------
        max_order : int
            max order of energy exchange iterations.
        sound_source : SoundSource
            sound source with ``sound_power`` and ``position``
        receiver : Receiver
            receiver object with position.
        speed_of_sound : float, optional
            Speed of sound in m/s.
        sampling_rate : int, optional
            _description_, by default 1000

        Returns
        -------
        _type_
            _description_

        """
        energy_response = np.zeros((self.n_bins, self.E_n_samples))

        for i_source, source_patch in enumerate(self.patches):

            source_pos = source_patch.center
            receiver_pos = receiver.position

            cos_xi = np.abs(np.sum(
                source_patch.normal*np.abs(receiver_pos-source_pos))) / \
                np.linalg.norm(source_pos-receiver_pos)

            difference = receiver_pos-source_pos
            difference = pf.Coordinates(
                difference[0], difference[1], difference[2])
            difference.radius = 1
            R = np.linalg.norm(receiver_pos-source_pos)
            delay = int(R/speed_of_sound * sampling_rate)

            i_patch = self.directivity_receivers.find_nearest(
                difference)[0][0]

            for k in range(max_order+1):
                for i_frequency in range(self.n_bins):
                    energy = self.E_matrix[
                        i_frequency, k, i_source, :, i_patch]
                    delayed_energy = _add_delay(energy, delay, axis=-1)

                    # Equation 20
                    factor = cos_xi * (np.exp(
                        -self.sound_attenuation_factor[i_frequency]*R)) / (
                            np.pi * R**2)
                    if factor <0:
                        print(factor)
                    receiver_energy = delayed_energy * factor
                    energy_response[i_frequency, ...] += receiver_energy

        return energy_response


def _init_energy_exchange(
            dl, dm, dn, dd_l, dd_m, S_x, S_y, S_z,
            sound_power, absorption, distance, attenuation, n_bins):
    half_l = dd_l/2
    half_m = dd_m/2

    sin_phi_delta = (dl + half_l - S_x)/ (np.sqrt(np.square(
        dl+half_l-S_x) + np.square(dm-S_y) + np.square(dn-S_z)))

    test1 = (dl - half_l) <= S_x
    test2 = S_x <= (dl + half_l)
    k_phi = -1 if test1 and test2 else 1
    sin_phi = k_phi * (dl - half_l - S_x) / (np.sqrt(np.square(
        dl-half_l-S_x) + np.square(dm-S_y) + np.square(dn-S_z)))
    if (sin_phi_delta-sin_phi) < 1e-11:
        sin_phi *= -1

    plus  = np.arctan(np.abs((dm+half_m-S_y)/np.abs(S_z)))
    minus = np.arctan(np.abs((dm-half_m-S_y)/np.abs(S_z)))

    test1 = (dm - half_m) <= S_y
    test2 = S_y <= (dm + half_m)
    k_beta = -1 if test1 and test2 else 1

    beta = np.abs(plus-(k_beta*minus))

    # don't forget to add constants
    # constants are
    energies = np.zeros(n_bins)
    for i_frequency in range(n_bins):
        alpha = absorption[i_frequency]
        constant = sound_power * (1-alpha) * (
            np.exp(-attenuation[i_frequency]*distance))

        energy = constant * (
            np.abs(sin_phi_delta-sin_phi) ) * beta / (4*np.pi)
        energies[i_frequency] = energy
    return energies

def _add_delay(ir, delay_samples, axis=-1):
    """Add delay to impulse response.

    Parameters
    ----------
    ir : np.ndarray
        impulse response which should be shifted.
    delay_samples : int
        delay samples.
    axis : int, optional
        axis which should be rolled, by default -1

    Returns
    -------
    ir : np.ndarray
        shifted impulse response.

    """
    ## add delay
    if delay_samples > ir.shape[axis]:
        raise ValueError(
            'length of ir is longer then ir delay is '
            f'{delay_samples} > {ir.shape[-1]}')
    ir_delayed = np.roll(ir, delay_samples, axis=axis)
    return ir_delayed


def _calc_incident_direction(position, normal, up_vector, target_position):
    """Calculate the incident direction of a sound wave.

    Parameters
    ----------
    position : np.ndarray
        Position of the wall.
    normal : np.ndarray
        Normal vector of the wall.
    up_vector : np.ndarray
        Up vector of the wall.
    target_position : np.ndarray
        Position of the source or other wall.

    Returns
    -------
    pf.Coordinates
        Incident direction of the sound wave.

    """
    direction = np.array(target_position) - np.array(position)

    x_dash = np.cross(normal, up_vector)
    y_dash = np.array(up_vector)
    z_dash = -np.array(normal)

    w_x_local = np.dot(direction, x_dash)
    w_y_local = np.dot(direction, y_dash)
    w_z_local = np.dot(direction, z_dash)

    azimuth = np.arctan2(-w_x_local, -w_z_local)
    w_y_local_normalized = w_y_local / np.linalg.norm(
        [w_x_local, w_y_local, w_z_local])
    elevation = np.arcsin(w_y_local_normalized)

    return pf.Coordinates.from_spherical_colatitude(
        azimuth, elevation, 1)


class RadiosityKang():
    """Radiosity object simulation."""

    speed_of_sound: float
    max_order_k: int
    sampling_rate: int
    patch_size: float
    ir_length_s: float
    patch_list: list[PatchesKang]

    def __init__(
            self, walls, patch_size, max_order_k, ir_length_s,
            speed_of_sound=346.18, sampling_rate=1000, absorption=0.1,
            source=None):
        """Create Radiosity Object for simulation.

        Parameters
        ----------
        walls : list[Polygon]
            list of patches
        patch_size : float
            maximal patchsize in meters.
        max_order_k : int
            max order of energy exchange iterations.
        ir_length_s : float
            length of ir in seconds.
        speed_of_sound : float, optional
            Speed of sound in m/s, by default 346.18 m/s
        sampling_rate : int, optional
            sampling rate of the Energy histogram, by default 1000 -> 1ms
        absorption: np.ndarray
            Absorption coefficient of this wall for each frequency bin
        source : SoundSource, optional
            Source object, by default None, can be added later.

        """
        self.speed_of_sound = speed_of_sound
        self.max_order_k = max_order_k
        self.sampling_rate = sampling_rate
        self.patch_size = patch_size
        self.ir_length_s = ir_length_s

        # A. Patch division
        patch_list = []
        n_patches = len(walls)
        for i, wall in enumerate(walls):
            index_list = [j for j in range(n_patches) if j != i]
            patches = wall if isinstance(wall, PatchesKang) else PatchesKang(
                wall, self.patch_size, index_list, i,absorption=absorption)
            patch_list.append(patches)

        self.patch_list = patch_list
        if source is not None:
            self.source = source

    def to_dict(self) -> dict:
        """Convert this object to dictionary. Used for read write."""
        is_source = hasattr(self, 'source')
        source = self.source if is_source else None
        return {
            'patch_size': self.patch_size,
            'max_order_k': self.max_order_k,
            'ir_length_s': self.ir_length_s,
            'speed_of_sound': self.speed_of_sound,
            'sampling_rate': self.sampling_rate,
            'patches': [patch.to_dict() for patch in self.patch_list],
            'source_position': source.position if is_source else None,
            'source_view': source.view if is_source else None,
            'source_up': source.up if is_source else None,
        }

    @classmethod
    def from_dict(cls, input_dict: dict):
        """Create an object from a dictionary. Used for read write."""
        patch_list = [
            PatchesKang.from_dict(patch) for patch in input_dict['patches']]
        source = None
        if input_dict is not None:
            source = SoundSource(
                input_dict['source_position'],
                input_dict['source_view'], input_dict['source_up'])
        obj = cls(
            patch_list, input_dict['patch_size'], input_dict['max_order_k'],
            input_dict['ir_length_s'],
            input_dict['speed_of_sound'], input_dict['sampling_rate'],
            source=source)
        return obj

    def run(self, source):
        """Execute the radiosity algorithm."""
        self.source = source
        # B. First-order patch sources
        for patches in self.patch_list:
            patches.init_energy_exchange(
                self.max_order_k, self.ir_length_s, source,
                sampling_rate=self.sampling_rate,
                speed_of_sound=self.speed_of_sound)

        # C. Form factors
        if len(self.patch_list) > 1:
            for patches in self.patch_list:
                patches.calculate_form_factor(self.patch_list)

        # D. Energy exchange between patches
        if len(self.patch_list) > 1:
            for k in range(1, self.max_order_k+1):
                for patches in self.patch_list:
                    patches.calculate_energy_exchange(
                        self.patch_list, k, speed_of_sound=self.speed_of_sound,
                        E_sampling_rate=self.sampling_rate)

    def energy_at_receiver(
            self, receiver, max_order_k=None, ignore_direct=False):
        """Return the energetic impulse response at the receiver."""
        ir = 0
        if max_order_k is None:
            max_order_k = self.max_order_k
        M_value = self.patch_list[0].sound_attenuation_factor
        for patches in self.patch_list:
            ir += patches.energy_at_receiver(
                max_order_k, receiver,
                speed_of_sound=self.speed_of_sound,
                sampling_rate=self.sampling_rate)
        if not ignore_direct:
            r = np.sqrt(np.sum((receiver.position-self.source.position)**2))
            direct_sound = (1/(4 * np.pi * np.square(r))) * np.exp(-M_value*r)
            delay_dir = int(r/self.speed_of_sound*self.sampling_rate)
            ir[:, delay_dir] += direct_sound

        return ir

    def write(self, filename, compress=True):
        """Write the object to a far file."""
        pf.io.write(filename, compress=compress, **self.to_dict())

    @classmethod
    def from_read(cls, filename):
        """Read the object to a far file."""
        data = pf.io.read(filename)
        return cls.from_dict(data)


class DirectionalRadiosityKang():
    """Radiosity object for directional scattering coefficients."""

    def __init__(
            self, polygon_list, patch_size, max_order_k, ir_length_s,
            sofa_path, speed_of_sound=346.18, sampling_rate=1000, source=None):
        """Create a Radiosity object for directional scattering coefficients.

        Parameters
        ----------
        polygon_list : list[PatchesDirectional]
            list of patches
        patch_size : float
            maximal patchsize in meters.
        max_order_k : int
            max order of energy exchange iterations.
        ir_length_s : float
            length of ir in seconds.
        sofa_path : Path, string, list of Path, list of string
            path of directional scattering coefficients or list of path
            for each Patch.
        speed_of_sound : float, optional
            Speed of sound in m/s, by default 346.18 m/s
        sampling_rate : int, optional
            sampling rate of the Energy histogram, by default 1000 -> 1ms
        source : SoundSource, optional
            Source object, by default None, can be added later.

        """
        self.speed_of_sound = speed_of_sound
        self.max_order_k = max_order_k
        self.sampling_rate = sampling_rate
        self.patch_size = patch_size
        self.ir_length_s = ir_length_s

        # A. Patch division
        patch_list = []
        n_points = len(polygon_list)
        for i in range(n_points):
            index_list = [j for j in range(n_points) if j != i]
            path = sofa_path[i] if isinstance(sofa_path, list) else sofa_path
            patches = polygon_list[i] if isinstance(
                polygon_list[i], PatchesDirectionalKang) else \
                    PatchesDirectionalKang.from_sofa(
                    polygon_list[i], self.patch_size, index_list, i, path)
            patch_list.append(patches)

        self.patch_list = patch_list
        n_bins = self.patch_list[0].n_bins
        for patches in self.patch_list:
            assert patches.n_bins == n_bins, \
                'Number of bins is not the same for all patches. ' + \
                f'{patches.n_bins} != {n_bins}'

        if source is not None:
            self.source = source

    def write(self, filename, compress=True):
        """Write the object to a far file."""
        pf.io.write(filename, compress=compress, **self.to_dict())

    @classmethod
    def from_read(cls, filename):
        """Read the object to a far file."""
        data = pf.io.read(filename)
        return cls.from_dict(data)

    def to_dict(self) -> dict:
        """Convert this object to dictionary. Used for read write."""
        is_source = hasattr(self, 'source')
        source = self.source if is_source else None
        return {
            'patch_size': self.patch_size,
            'max_order_k': self.max_order_k,
            'ir_length_s': self.ir_length_s,
            'speed_of_sound': self.speed_of_sound,
            'sampling_rate': self.sampling_rate,
            'patches': [patch.to_dict() for patch in self.patch_list],
            'source_position': source.position if is_source else None,
            'source_view': source.view if is_source else None,
            'source_up': source.up if is_source else None,
        }

    @classmethod
    def from_dict(cls, input_dict: dict):
        """Create an object from a dictionary. Used for read write."""
        patch_list = [
            PatchesDirectionalKang.from_dict(patch) for patch in input_dict[
                'patches']]
        source = None
        if input_dict['source_position'] is not None:
            source = SoundSource(
                input_dict['source_position'],
                input_dict['source_view'], input_dict['source_up'])
        obj = cls(
            patch_list, input_dict['patch_size'], input_dict['max_order_k'],
            input_dict['ir_length_s'],
            sofa_path=None,
            speed_of_sound=input_dict['speed_of_sound'],
            sampling_rate=input_dict['sampling_rate'],
            source=source,
            )
        return obj

    def run(self, source):
        """Execute the radiosity algorithm."""
        self.source = source
        # B. First-order patch sources
        for patches in self.patch_list:
            patches.init_energy_exchange(
                self.max_order_k, self.ir_length_s, source,
                sampling_rate=self.sampling_rate,
                speed_of_sound=self.speed_of_sound)

        # C. Form factors
        if len(self.patch_list) > 1:
            for patches in self.patch_list:
                patches.calculate_form_factor(self.patch_list)

        # D. Energy exchange between patches
        if len(self.patch_list) > 1:
            for k in range(1, self.max_order_k+1):
                for patches in self.patch_list:
                    patches.calculate_energy_exchange(
                        self.patch_list, k, speed_of_sound=self.speed_of_sound,
                        E_sampling_rate=self.sampling_rate)

    def energy_at_receiver(self, receiver, order_k=None):
        """Return the energetic impulse response at the receiver."""
        ir = 0
        if order_k is None:
            order_k = self.max_order_k
        M = self.patch_list[0].sound_attenuation_factor
        for patches in self.patch_list:
            ir += patches.energy_at_receiver(
                order_k, receiver,
                speed_of_sound=self.speed_of_sound,
                sampling_rate=self.sampling_rate)
        r = np.sqrt(np.sum((receiver.position-self.source.position)**2))
        direct_sound = (1/(4 * np.pi * np.square(r))) * np.exp(-M*r)
        delay_dir = int(r/self.speed_of_sound*self.sampling_rate)
        ir[:, delay_dir] += direct_sound

        return ir



"""Test the radiosity.Radiosity module."""
import os

import numpy as np
import numpy.testing as npt
import pyfar as pf
import pytest
import sparrowpy as sp
import sparrowpy.geometry as geo
from sparrowpy.sound_object import Receiver, SoundSource


create_reference_files = False


def test_small_room_and_shift():
    """Test if the results changes for shifted walls."""
    X = 5
    Y = 6
    Z = 4
    r_x = 3
    r_y = 2
    r_z = 3
    patch_size = 1
    ir_length_s = 1
    sampling_rate = 1
    max_order_k = 10
    speed_of_sound = 343
    irs_new = []
    E_matrix = []
    form_factors = []
    for i in range(2):
        walls = sp.testing.shoebox_room_stub(X, Y, Z)
        if i == 0:
            delta_x = 0.
            delta_y = 0.
            delta_z = 0.
        elif i == 1:
            delta_x = 2.
            delta_y = 4.
            delta_z = 5.

        receiver_pos = [r_x+delta_x, r_y+delta_y, r_z+delta_z]
        for wall in walls:
            wall.pts += np.array([delta_x, delta_y, delta_z])

        # create geometry
        source = sp.geometry.SoundSource(
            [2+delta_x, 2+delta_y, 2+delta_z], [0, 1, 0], [0, 0, 1])

        ## new approach
        radi = sp.RadiosityKang(
            walls, patch_size, max_order_k, ir_length_s,
            speed_of_sound=speed_of_sound, sampling_rate=sampling_rate,
            absorption=0.1)

        # run simulation
        radi.run(source)

        E_matrix.append(np.concatenate([
            radi.patch_list[i].E_matrix for i in range(6)], axis=-2))
        form_factors.append([
            radi.patch_list[i].form_factors for i in range(6)])

        # test energy at receiver
        receiver = sp.geometry.Receiver(receiver_pos, [0, 1, 0], [0, 0, 1])
        irs_new.append(radi.energy_at_receiver(receiver, ignore_direct=True))

    # compare E_matrix
    E_matrix = np.array(E_matrix)
    npt.assert_array_equal(E_matrix[0], E_matrix[1])

    # compare form factors
    for i in range(6):
        npt.assert_array_equal(form_factors[0][i], form_factors[1][i])

    # rotate all walls
    irs_new = np.array(irs_new).squeeze()
    npt.assert_array_equal(irs_new, irs_new[0])


def test_small_room_and_rotate():
    """Test if the results changes for rotated walls."""
    X = 5
    Y = 3
    Z = 4
    r_x = 1
    r_y = 2
    r_z = 1.5
    patch_size = 1
    ir_length_s = 1
    sampling_rate = 1
    max_order_k = 10
    speed_of_sound = 343
    irs_new = []
    E_matrix = []
    E_matrix_sum = []
    for i in range(6):
        if i == 0:
            walls = sp.testing.shoebox_room_stub(X, Y, Z)
            receiver_pos = [r_x, r_y, r_z]
        elif i == 1:
            walls = sp.testing.shoebox_room_stub(X, Z, Y)
            receiver_pos = [r_x, r_z, r_y]
        elif i == 2:
            walls = sp.testing.shoebox_room_stub(Y, Z, X)
            receiver_pos = [r_y, r_z, r_x]
        elif i == 3:
            walls = sp.testing.shoebox_room_stub(Y, X, Z)
            receiver_pos = [r_y, r_x, r_z]
        elif i == 4:
            walls = sp.testing.shoebox_room_stub(Z, X, Y)
            receiver_pos = [r_z, r_x, r_y]
        else:
            walls = sp.testing.shoebox_room_stub(Z, Y, X)
            receiver_pos = [r_z, r_y, r_x]

        # create geometry
        source = sp.geometry.SoundSource([2, 2, 2], [0, 1, 0], [0, 0, 1])

        ## new approach
        radi = sp.RadiosityKang(
            walls, patch_size, max_order_k, ir_length_s,
            speed_of_sound=speed_of_sound, sampling_rate=sampling_rate,
            absorption=0.1)

        # run simulation
        radi.run(source)

        E_matrix.append(np.concatenate([
            radi.patch_list[i].E_matrix for i in range(6)], axis=-2))
        E_matrix_sum.append(
            [radi.patch_list[i].E_matrix.sum() for i in range(6)])
        # test energy at receiver
        receiver = sp.geometry.Receiver(receiver_pos, [0, 1, 0], [0, 0, 1])
        irs_new.append(radi.energy_at_receiver(receiver, ignore_direct=True))

    # rotate all walls
    irs_new = np.array(irs_new).squeeze()
    npt.assert_array_almost_equal(irs_new, irs_new[0], decimal=4)


def test_small_room_and_rotate_init_energy():
    """Test if the results changes for rotated walls."""
    X = 5
    Y = 6
    Z = 4
    patch_size = 1
    ir_length_s = 1
    sampling_rate = 1
    max_order_k = 0
    speed_of_sound = 343
    E_matrix = []
    E_matrix_sum = []
    for i in range(6):
        if i == 0:
            walls = sp.testing.shoebox_room_stub(X, Y, Z)
        elif i == 1:
            walls = sp.testing.shoebox_room_stub(X, Z, Y)
        elif i == 2:
            walls = sp.testing.shoebox_room_stub(Y, Z, X)
        elif i == 3:
            walls = sp.testing.shoebox_room_stub(Y, X, Z)
        elif i == 4:
            walls = sp.testing.shoebox_room_stub(Z, X, Y)
        else:
            walls = sp.testing.shoebox_room_stub(Z, Y, X)

        # create geometry
        source = sp.geometry.SoundSource([2, 2, 2], [0, 1, 0], [0, 0, 1])

        ## new approach
        radi = sp.RadiosityKang(
            walls, patch_size, max_order_k, ir_length_s,
            speed_of_sound=speed_of_sound, sampling_rate=sampling_rate,
            absorption=0.1)

        # run init energy
        # B. First-order patch
        for patches in radi.patch_list:
            patches.init_energy_exchange(
                radi.max_order_k, radi.ir_length_s, source,
                sampling_rate=radi.sampling_rate,
                speed_of_sound=speed_of_sound)

        E_matrix.append(np.concatenate([
            radi.patch_list[i].E_matrix for i in range(6)], axis=-2))
        E_matrix_sum.append(
            [radi.patch_list[i].E_matrix.sum() for i in range(6)])

    # rotate all walls
    E_matrix_sum = np.array(E_matrix_sum)
    matrix = np.zeros(E_matrix_sum.shape, dtype=int)
    reference_energy = E_matrix_sum[:, 2] # formula was given for the ground
    for i, ref in enumerate(reference_energy):
        matrix[np.abs(E_matrix_sum-ref)<1e-10] = i+1
    E_matrix = np.array(E_matrix)
    E_matrix = np.sum(E_matrix[:, 0, 0, :, 0], axis=-1)
    npt.assert_array_almost_equal(E_matrix, E_matrix[0], decimal=4)


@pytest.mark.parametrize('patch_size', [1, 0.5])
def test_cube_and_rotate_init_energy(patch_size):
    """Test if the results changes for rotated walls."""
    ir_length_s = 1
    sampling_rate = 1
    max_order_k = 0
    speed_of_sound = 343
    walls = sp.testing.shoebox_room_stub(1, 1, 1)
    # create geometry
    source = sp.geometry.SoundSource([0.5, 0.5, 0.5], [0, 1, 0], [0, 0, 1])

    ## new approach
    radi = sp.RadiosityKang(
        walls, patch_size, max_order_k, ir_length_s,
        speed_of_sound=speed_of_sound, sampling_rate=sampling_rate,
        absorption=0.1)

    # run init energy
    # B. First-order patch sources
    for patches in radi.patch_list:
        print(patches.normal)
        patches.init_energy_exchange(
            radi.max_order_k, radi.ir_length_s, source,
            sampling_rate=radi.sampling_rate, speed_of_sound=speed_of_sound)

    E_matrix= np.concatenate([
        radi.patch_list[i].E_matrix for i in range(6)], axis=-2)

    # rotate all walls
    assert E_matrix.flatten()[0] > 0
    npt.assert_array_equal(E_matrix, E_matrix.flatten()[0])


@pytest.mark.parametrize('max_order_k', [2, 3])
def test_radiosity_reference(max_order_k):
    """Test if the results changes."""
    X = 10
    Y = 20
    Z = 15
    patch_size = 5
    ir_length_s = 5
    sampling_rate = 50
    # create geometry
    ground = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, Y, 0], [0, Y, 0]], [1, 0, 0], [0, 0, 1])
    A_wall = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, 0, Z], [0, 0, Z]], [1, 0, 0], [0, 1, 0])
    B_wall = geo.Polygon(
        [[0, Y, 0], [X, Y, 0], [X, Y, Z], [0, Y, Z]], [1, 0, 0], [0, -1, 0])
    source = SoundSource([10, 6, 1], [0, 1, 0], [0, 0, 1])

    ## new approach
    radi = sp.RadiosityKang(
        [ground, A_wall, B_wall], patch_size, max_order_k, ir_length_s,
        speed_of_sound=343, sampling_rate=sampling_rate)

    radi.run(source)

    receiver = Receiver([20, 2, 1], [0, 1, 0], [0, 0, 1])
    irs_new = radi.energy_at_receiver(receiver)
    signal = pf.Signal(irs_new, sampling_rate)
    signal.time /= np.max(np.abs(signal.time))

    test_path = os.path.join(
        os.path.dirname(__file__), 'test_data')

    # write test file
    if create_reference_files:
        pf.io.write(
            os.path.join(
                test_path,
                f'simulation_X{X}_k{max_order_k}_{patch_size}m.far'),
            signal=signal)

    result = pf.io.read(
        os.path.join(
            test_path,
            f'simulation_X{X}_k{max_order_k}_{patch_size}m.far'))

    npt.assert_almost_equal(
        result['signal'].time[0, ...], signal.time[0, ...], decimal=2)
    npt.assert_almost_equal(result['signal'].times, signal.times)


@pytest.mark.parametrize('max_order_k', [2, 3])
def test_radiosity_reference_with_read_write(max_order_k, tmpdir):
    """Test if the results changes."""
    X = 10
    Y = 20
    Z = 15
    patch_size = 5
    ir_length_s = 5
    sampling_rate = 50
    # create geometry
    ground = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, Y, 0], [0, Y, 0]], [1, 0, 0], [0, 0, 1])
    A_wall = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, 0, Z], [0, 0, Z]], [1, 0, 0], [0, 1, 0])
    B_wall = geo.Polygon(
        [[0, Y, 0], [X, Y, 0], [X, Y, Z], [0, Y, Z]], [1, 0, 0], [0, -1, 0])
    source = SoundSource([10, 6, 1], [0, 1, 0], [0, 0, 1])

    ## new approach
    radi = sp.RadiosityKang(
        [ground, A_wall, B_wall], patch_size, max_order_k, ir_length_s,
        speed_of_sound=343, sampling_rate=sampling_rate)

    radi.run(source)
    path = os.path.join(tmpdir, 'radiosity.far')
    radi.write(path)
    radi = sp.RadiosityKang.from_read(path)
    receiver = Receiver([20, 2, 1], [0, 1, 0], [0, 0, 1])
    irs_new = radi.energy_at_receiver(receiver)
    signal = pf.Signal(irs_new, sampling_rate)
    signal.time /= np.max(np.abs(signal.time))

    test_path = os.path.join(
        os.path.dirname(__file__), 'test_data')

    # write test file
    result = pf.io.read(
        os.path.join(
            test_path,
            f'simulation_X{X}_k{max_order_k}_{patch_size}m.far'))

    npt.assert_almost_equal(
        result['signal'].time[0, ...], signal.time[0, ...], decimal=2)
    npt.assert_almost_equal(result['signal'].times, signal.times)


@pytest.mark.parametrize('patch_size', [
    0.5,
    1,
    ])
@pytest.mark.parametrize('i_wall', [0, 1, 2, 3, 4, 5])
def test_init_energy_exchange_normal(sample_walls, patch_size, i_wall):
    """Test init energy exchange."""
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data',
        f'reference_matrix_directional_patch_size{patch_size}.far')
    patches = sp.PatchesKang(sample_walls[i_wall], patch_size, [], 0)
    max_order_k = 3
    ir_length_s = 5
    source = SoundSource([0.5, 0.5, 0.5], [0, 1, 0], [0, 0, 1])
    patches.init_energy_exchange(
        max_order_k, ir_length_s, source, 1000, 346.18)
    if create_reference_files and sample_walls[i_wall] == sample_walls[0]:
        pf.io.write(path_sofa, E_matrix=patches.E_matrix)
    data = pf.io.read(path_sofa)
    assert np.sum(patches.E_matrix>0) > 0
    npt.assert_almost_equal(
        data['E_matrix'], patches.E_matrix, decimal=4)


@pytest.mark.parametrize('parallel_walls', [
    [0, 1], [1, 0], [2, 3], [3, 2], [4, 5], [5, 4],
    ])
@pytest.mark.parametrize('patch_size', [
    0.5,
    1,
    ])
def test_calc_form_factor_parallel(sample_walls, parallel_walls, patch_size):
    """Test form factor calculation for parallel walls."""
    wall_source = sample_walls[parallel_walls[0]]
    wall_receiver = sample_walls[parallel_walls[1]]
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data',
        f'reference_form_factor_parallel_size{patch_size}.far')
    patch_1 = sp.PatchesKang(wall_source, patch_size, [1], 0)
    patch_2 = sp.PatchesKang(wall_receiver, patch_size, [0], 1)
    patches = [patch_1, patch_2]
    patch_1.calculate_form_factor(patches)
    if create_reference_files and (
            parallel_walls[0] == 0) and (parallel_walls[1] == 1):
        pf.io.write(path_sofa, form_factor=patch_1.form_factors)
    data = pf.io.read(path_sofa)
    npt.assert_almost_equal(data['form_factor'], patch_1.form_factors)

@pytest.mark.parametrize('perpendicular_walls', [
    [0, 2], [0, 3], [0, 4], [0, 5],
    [1, 2], [1, 3], [1, 4], [1, 5],
    [2, 0], [2, 1], [2, 4], [2, 5],
    [3, 0], [3, 1], [3, 4], [3, 5],
    [4, 0], [4, 1], [4, 2], [4, 3],
    [5, 0], [5, 1], [5, 2], [5, 3],
    ])
@pytest.mark.parametrize('patch_size', [
    0.5,
    1,
    ])
def test_calc_form_factor_perpendicular(
        sample_walls, perpendicular_walls, patch_size):
    """Test form factor calculation for perpendicular walls."""
    wall_source = sample_walls[perpendicular_walls[0]]
    wall_receiver = sample_walls[perpendicular_walls[1]]
    idx_sort = [0, 0] if patch_size == 1 else perpendicular_walls
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data',
        f'reference_form_factor_perpendicular_{idx_sort[0]}_{idx_sort[1]}'
        f'_size{patch_size}.far')
    patch_1 = sp.PatchesKang(wall_source, patch_size, [1], 0)
    patch_2 = sp.PatchesKang(wall_receiver, patch_size, [0], 1)
    patches = [patch_1, patch_2]
    patch_1.calculate_form_factor(patches)
    if create_reference_files:
        pf.io.write(path_sofa, form_factor=patch_1.form_factors)
    data = pf.io.read(path_sofa)
    npt.assert_almost_equal(data['form_factor'], patch_1.form_factors)


@pytest.mark.parametrize('perpendicular_walls', [
    [0, 2], [0, 3], [0, 4], [0, 5],
    [1, 2], [1, 3], [1, 4], [1, 5],
    [2, 0], [2, 1], [2, 4], [2, 5],
    [3, 0], [3, 1], [3, 4], [3, 5],
    [4, 0], [4, 1], [4, 2], [4, 3],
    [5, 0], [5, 1], [5, 2], [5, 3],
    ])
@pytest.mark.parametrize('patch_size', [
    0.5,
    ])
def test_calc_form_factor_perpendicular_distance(
        sample_walls, perpendicular_walls, patch_size):
    """Test form factor calculation for perpendicular walls."""
    wall_source = sample_walls[perpendicular_walls[0]]
    wall_receiver = sample_walls[perpendicular_walls[1]]
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data',
        f'reference_form_factor_perpendicular_size{patch_size}.far')
    patch_1 = sp.PatchesKang(wall_source, patch_size, [1], 0)
    patch_2 = sp.PatchesKang(wall_receiver, patch_size, [0], 1)
    patches = [patch_1, patch_2]
    patch_1.calculate_form_factor(patches)
    ff_sort = patch_1.form_factors.flatten()
    ff_sort.sort()
    if create_reference_files and (
            perpendicular_walls[0] == 0) and (perpendicular_walls[1] == 2):
        pf.io.write(path_sofa, form_factor=ff_sort)
    data = pf.io.read(path_sofa)
    npt.assert_almost_equal(data['form_factor'], ff_sort)


@pytest.mark.parametrize('perpendicular_walls', [
    [0, 2], [2, 0],
    ])
@pytest.mark.parametrize('patch_size', [
    0.5,
    1,
    ])
def test_energy_exchange(
        sample_walls, perpendicular_walls, patch_size):
    """Test energy exchange."""
    max_order_k=3
    ir_length_s=5
    wall_source = sample_walls[perpendicular_walls[0]]
    wall_receiver = sample_walls[perpendicular_walls[1]]
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data',
        f'reference_energy_exchange_size{patch_size}.far')
    patch_1 = sp.PatchesKang(
        wall_source, patch_size, [1], 0)
    patch_2 = sp.PatchesKang(
        wall_receiver, patch_size, [0], 1)
    source = SoundSource([0.5, 0.5, 0.5], [0, 1, 0], [0, 0, 1])
    patches = [patch_1, patch_2]
    patch_1.calculate_form_factor(patches)
    patch_2.calculate_form_factor(patches)
    patch_1.init_energy_exchange(
        max_order_k, ir_length_s, source, 1000, 346.18)
    patch_2.init_energy_exchange(
        max_order_k, ir_length_s, source, 1000, 346.18)
    for k in range(1, max_order_k+1):
        patch_1.calculate_energy_exchange(
            patches, k, speed_of_sound=346.18, E_sampling_rate=1000)
        patch_2.calculate_energy_exchange(
            patches, k, speed_of_sound=346.18, E_sampling_rate=1000)

    if create_reference_files and (
            perpendicular_walls[0] == 0) and (perpendicular_walls[1] == 2):
        pf.io.write(path_sofa, E_matrix=patch_1.E_matrix)
    data = pf.io.read(path_sofa)

    assert np.sum(patch_1.E_matrix>0) > 0
    npt.assert_almost_equal(
        10*np.log10(data['E_matrix']),
        10*np.log10(patch_1.E_matrix), decimal=1)


def test_Patch_to_from_dict(sample_walls):
    """Test Patches from dict."""
    perpendicular_walls = [0, 2]
    patch_size = 0.5
    max_order_k=3
    ir_length_s=5
    wall_source = sample_walls[perpendicular_walls[0]]
    patch_1 = sp.PatchesKang(
        wall_source, patch_size, [1], 0)
    source = SoundSource([0.5, 0.5, 0.5], [0, 1, 0], [0, 0, 1])
    patch_1.init_energy_exchange(
        max_order_k, ir_length_s, source, 1000, 346.18)
    reconstructed_patch = sp.PatchesKang.from_dict(
        patch_1.to_dict())
    npt.assert_array_equal(reconstructed_patch.E_matrix, patch_1.E_matrix)
    npt.assert_array_equal(
        reconstructed_patch.form_factors, patch_1.form_factors)
    npt.assert_array_equal(reconstructed_patch.pts, patch_1.pts)
    npt.assert_array_equal(reconstructed_patch.up_vector, patch_1.up_vector)
    npt.assert_array_equal(reconstructed_patch.normal, patch_1.normal)
    assert reconstructed_patch.max_size == patch_1.max_size
    assert reconstructed_patch.other_wall_ids == patch_1.other_wall_ids
    assert reconstructed_patch.wall_id == patch_1.wall_id
    npt.assert_array_equal(reconstructed_patch.scattering, patch_1.scattering)
    npt.assert_array_equal(reconstructed_patch.absorption, patch_1.absorption)
    npt.assert_array_equal(
        reconstructed_patch.sound_attenuation_factor,
        patch_1.sound_attenuation_factor)


def test_radiosity_to_from_dict():
    """Test Radiosity the results changes for to and from dict."""
    max_order_k = 3
    patch_size = 5
    X = 10
    Y = 20
    Z = 15
    ir_length_s = 5
    sampling_rate = 50
    # create geometry
    ground = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, Y, 0], [0, Y, 0]], [1, 0, 0], [0, 0, 1])
    A_wall = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, 0, Z], [0, 0, Z]], [1, 0, 0], [0, 1, 0])
    B_wall = geo.Polygon(
        [[0, Y, 0], [X, Y, 0], [X, Y, Z], [0, Y, Z]], [1, 0, 0], [0, -1, 0])
    source = SoundSource([10, 6, 1], [0, 1, 0], [0, 0, 1])

    ## new approach
    radi = sp.RadiosityKang(
        [ground, A_wall, B_wall], patch_size, max_order_k, ir_length_s,
        speed_of_sound=343, sampling_rate=sampling_rate)

    radi.run(source)
    radi_dict = radi.to_dict()
    radi_reconstructed = sp.RadiosityKang.from_dict(radi_dict)

    # test
    assert isinstance(radi_reconstructed, sp.RadiosityKang)
    assert radi_reconstructed.patch_size == radi.patch_size
    assert radi_reconstructed.max_order_k == radi.max_order_k
    assert radi_reconstructed.ir_length_s == radi.ir_length_s
    assert radi_reconstructed.speed_of_sound == radi.speed_of_sound
    assert radi_reconstructed.sampling_rate == radi.sampling_rate
    assert len(radi_reconstructed.patch_list) == len(radi.patch_list)
    for patch, patch_reconstructed in zip(
            radi.patch_list, radi_reconstructed.patch_list, strict=True):
        np.testing.assert_array_equal(
            patch.pts, patch_reconstructed.pts)
        np.testing.assert_array_equal(
            patch.up_vector, patch_reconstructed.up_vector)
        np.testing.assert_array_equal(
            patch._normal, patch_reconstructed._normal)
        assert patch.max_size == patch_reconstructed.max_size
        assert patch.wall_id == patch_reconstructed.wall_id
        np.testing.assert_array_equal(
            patch.scattering, patch_reconstructed.scattering)
        np.testing.assert_array_equal(
            patch.absorption, patch_reconstructed.absorption)
        np.testing.assert_array_equal(
            patch.sound_attenuation_factor,
            patch_reconstructed.sound_attenuation_factor)
        np.testing.assert_array_equal(
            patch.E_matrix, patch_reconstructed.E_matrix)


def test_radiosity_read_write(tmpdir):
    """Test Radiosity the results changes for read and write."""
    max_order_k = 3
    patch_size = 5
    X = 10
    Y = 20
    Z = 15
    ir_length_s = 5
    sampling_rate = 50
    # create geometry
    ground = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, Y, 0], [0, Y, 0]], [1, 0, 0], [0, 0, 1])
    A_wall = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, 0, Z], [0, 0, Z]], [1, 0, 0], [0, 1, 0])
    B_wall = geo.Polygon(
        [[0, Y, 0], [X, Y, 0], [X, Y, Z], [0, Y, Z]], [1, 0, 0], [0, -1, 0])
    source = SoundSource([10, 6, 1], [0, 1, 0], [0, 0, 1])

    ## new approach
    radi = sp.RadiosityKang(
        [ground, A_wall, B_wall], patch_size, max_order_k, ir_length_s,
        speed_of_sound=343, sampling_rate=sampling_rate)

    radi.run(source)

    json_path = os.path.join(tmpdir, 'radi.far')
    radi.write(json_path)
    radi_reconstructed = sp.RadiosityKang.from_read(json_path)

    # test
    assert isinstance(radi_reconstructed, sp.RadiosityKang)
    assert radi_reconstructed.patch_size == radi.patch_size
    assert radi_reconstructed.max_order_k == radi.max_order_k
    assert radi_reconstructed.ir_length_s == radi.ir_length_s
    assert radi_reconstructed.speed_of_sound == radi.speed_of_sound
    assert radi_reconstructed.sampling_rate == radi.sampling_rate
    assert len(radi_reconstructed.patch_list) == len(radi.patch_list)
    for patch, patch_reconstructed in zip(
            radi.patch_list, radi_reconstructed.patch_list, strict=True):
        np.testing.assert_array_equal(
            patch.pts, patch_reconstructed.pts)
        np.testing.assert_array_equal(
            patch.up_vector, patch_reconstructed.up_vector)
        np.testing.assert_array_equal(
            patch._normal, patch_reconstructed._normal)
        assert patch.max_size == patch_reconstructed.max_size
        assert patch.wall_id == patch_reconstructed.wall_id
        np.testing.assert_array_equal(
            patch.scattering, patch_reconstructed.scattering)
        np.testing.assert_array_equal(
            patch.absorption, patch_reconstructed.absorption)
        np.testing.assert_array_equal(
            patch.sound_attenuation_factor,
            patch_reconstructed.sound_attenuation_factor)
        np.testing.assert_array_equal(
            patch.E_matrix, patch_reconstructed.E_matrix)


@pytest.mark.parametrize('patch_size', [1, 0.5, 0.2, 1/3])
def test_init_energy_larger_0(patch_size):
    """Test if the results changes for rotated walls."""
    ir_length_s = 1
    sampling_rate = 1
    max_order_k = 0
    speed_of_sound = 343
    walls = sp.testing.shoebox_room_stub(1, 1, 1)
    # create geometry
    source = sp.geometry.SoundSource([0.5, 0.5, 0.5], [0, 1, 0], [0, 0, 1])

    ## new approach
    radi = sp.RadiosityKang(
        walls, patch_size, max_order_k, ir_length_s,
        speed_of_sound=speed_of_sound, sampling_rate=sampling_rate,
        absorption=0.1)

    # run init energy
    # B. First-order patch sources
    for patches in radi.patch_list:
        patches.init_energy_exchange(
            radi.max_order_k, radi.ir_length_s, source,
            sampling_rate=radi.sampling_rate, speed_of_sound=speed_of_sound)

    E_matrix= np.concatenate([
        radi.patch_list[i].E_matrix for i in range(6)], axis=-2)

    # rotate all walls
    assert E_matrix.flatten()[0] > 0




# -*- coding: utf-8 -*-

"""Top-level package for sparrowpy."""

__author__ = """The sparrowpy developers"""
__email__ = ''
__version__ = '1.0.1'


from . import geometry
from . import sound_object
from .classes.RadiosityFast import DirectionalRadiosityFast
from .classes.RadiosityKang import (
    RadiosityKang,
    DirectionalRadiosityKang,
    PatchesKang,
    PatchesDirectionalKang)
from . import testing
from . import form_factor
from . import brdf
from . import utils


__all__ = [
    'RadiosityKang',
    'DirectionalRadiosityKang',
    'PatchesKang',
    'PatchesDirectionalKang',
    'geometry',
    'sound_object',
    'DirectionalRadiosityFast',
    'form_factor',
    'testing',
    'brdf',
    'utils',
]



"""Contain functions to create BRDFs from scattering coefficients."""
import numpy as np
import sofar as sf
import pyfar as pf
import sparrowpy


def create_from_scattering(
        source_directions,
        receiver_directions,
        scattering_coefficient,
        absorption_coefficient=None,
        file_path=None,
        ):
    r"""Create the BRDF from a scattering coefficient and write to SOFA file.

    The scattering coefficient is assumed to be anisotropic and based on [#]_.
    The BRDF is discretized as follows:

    .. math::
        \rho(\mathbf{\Omega_i}, \mathbf{\Omega_o}) =
        \frac{(1-s)(1-\alpha)}{\mathbf{\Omega_i} \cdot \mathbf{n}}
        \frac{1}{w_o} \delta(\mathbf{\Omega_i}-M(\mathbf{\Omega_o})) +
        \frac{s(1-\alpha)}{\pi}

    where:
        - :math:`\mathbf{\Omega_i}` and :math:`\mathbf{\Omega_o}` are the
          incident and outgoing directions, respectively.
        - :math:`s` is the scattering coefficient.
        - :math:`\alpha` is the absorption coefficient.
        - :math:`\mathbf{n}` is the normal vector of the surface.
        - :math:`\delta` is the Dirac delta function.
        - :math:`w_o` is weighting factor of the outgoing angular sector
          (unit sphere).
        - :math:`M` s the mirror reflection transformation
          :math:`M(\theta, \phi)=M(\theta, \pi-\phi)`.

    Note that the weights doesn't need to be normalized,
    they get scaled as required.

    Parameters
    ----------
    source_directions : :py:class:`~pyfar.classes.coordinates.Coordinates`
        source directions for the BRDF, should contain weights.
        cshape of data should be (n_sources)
    receiver_directions : :py:class:`~pyfar.classes.coordinates.Coordinates`
        receiver directions for the BRDF, should contain weights.
        cshape of data should be (n_receivers)
    scattering_coefficient : :py:class:`~pyfar.classes.audio.FrequencyData`
        frequency dependent scattering coefficient data.
        cshape of data should be (1, ).
    absorption_coefficient : :py:class:`~pyfar.classes.audio.FrequencyData`
        frequency dependent absorption coefficient data, by default
        no absorption.
        cshape of data should be (1, ).
    file_path : string, path
        path where sofa file should be saved, by default no file is saved.

    Returns
    -------
    brdf : :py:class:`~pyfar.classes.audio.FrequencyData`
        BRDF data.

    References
    ----------
    .. [#]  S. Siltanen, T. Lokki, S. Kiminki, and L. Savioja, “The room
            acoustic rendering equation,” The Journal of the Acoustical
            Society of America, vol. 122, no. 3, pp. 1624-1635, 2007.

    Examples
    --------
    >>> import pyfar as pf
    >>> import sparrowpy as sp
    >>> import numpy as np
    >>> scattering_coefficient = pf.FrequencyData(0.5, [100])
    >>> directions = pf.samplings.sph_gaussian(sh_order=3)
    >>> directions = directions[directions.z > 0]
    >>> brdf = sp.brdf.create_from_scattering(
    ...     directions, directions,
    ...     scattering_coefficient)

    """
    if (
            not isinstance(scattering_coefficient, pf.FrequencyData) or
            not scattering_coefficient.cshape == (1,)):
        raise TypeError(
            'scattering_coefficient must be a pf.FrequencyData object'
            'with shape (1,)')
    if not isinstance(source_directions, pf.Coordinates):
        raise TypeError(
            'source_directions must be a pf.Coordinates object')
    if not isinstance(receiver_directions, pf.Coordinates):
        raise TypeError(
            'receiver_directions must be a pf.Coordinates object')
    if absorption_coefficient is None:
        absorption_coefficient = pf.FrequencyData(
            np.zeros_like(scattering_coefficient.frequencies),
            scattering_coefficient.frequencies)

    brdf = np.zeros((
        source_directions.csize, receiver_directions.csize,
        scattering_coefficient.n_bins))

    receiver_weights = receiver_directions.weights
    receiver_weights *= 2 * np.pi / np.sum(receiver_weights)
    scattering_flattened = np.real(scattering_coefficient.freq.flatten())
    image_source = source_directions.copy()
    image_source.azimuth += np.pi
    i_receiver = receiver_directions.find_nearest(image_source)[0][0]
    cos_factor = (np.cos(
            source_directions.colatitude[
                np.newaxis, ...]) * receiver_weights[..., np.newaxis])
    scattering_factor = 1 - scattering_flattened[np.newaxis, ...]
    brdf[:, :, :] += (
        scattering_flattened) / np.pi
    i_sources = np.arange(source_directions.csize)
    brdf[i_sources, i_receiver, :] += scattering_factor / cos_factor[
        i_sources, i_receiver, np.newaxis]

    brdf *= (1 - absorption_coefficient.freq.flatten())
    if file_path is not None:
        sofa = _create_sofa(
            pf.FrequencyData(brdf, scattering_coefficient.frequencies),
            source_directions,
            receiver_directions,
            history='constructed brdf based on scattering coefficients',
        )

        sf.write_sofa(file_path, sofa)
    return pf.FrequencyData(brdf, scattering_coefficient.frequencies)


def create_from_directional_scattering(
        source_directions,
        receiver_directions,
        directional_scattering,
        absorption_coefficient=None,
        file_path=None,
        ):
    r"""Create the BRDF from the directional scattering and write to SOFA file.

    The directional scattering coefficient is assumed to be anisotropic.
    The sum of the directional scattering coefficient has be equal to 1.
    Therefore the BRDF is calculated as follows:

    .. math::
        \rho(\mathbf{\Omega_i}, \mathbf{\Omega_o}) = \frac{(1-\alpha)}{
        (\mathbf{\Omega_o} \cdot \mathbf{n}) \cdot w_o} s_{d}(
        \mathbf{\Omega_i}, \mathbf{\Omega_o})

    where:
        - :math:`\mathbf{\Omega_i}` and :math:`\mathbf{\Omega_o}` are the
          incident and outgoing directions, respectively.
        - :math:`s_{d}` is the directional scattering coefficient [#]_.
        - :math:`\alpha` is the absorption coefficient.
        - :math:`\mathbf{n}` is the normal vector of the surface.
        - :math:`w_o` is weighting factor of the outgoing angular sector
          (unit sphere).

    Note that the weights doesn't need to be normalized,
    they get scaled as required.

    Parameters
    ----------
    source_directions : :py:class:`~pyfar.classes.coordinates.Coordinates`
        source directions for the BRDF, should contain weights.
        cshape of data should be (n_sources)
    receiver_directions : :py:class:`~pyfar.classes.coordinates.Coordinates`
        receiver directions for the BRDF, should contain weights.
        cshape of data should be (n_receivers)
    directional_scattering : :py:class:`~pyfar.classes.audio.FrequencyData`
        frequency dependent directional scattering coefficient data from [1]_.
        cshape of data should be (n_sources, n_receivers).
    absorption_coefficient : :py:class:`~pyfar.classes.audio.FrequencyData`
        frequency dependent absorption coefficient data, by default
        no absorption.
        cshape of data should be (1, ).
    file_path : string, path, optional
        path where sofa file should be saved, by default no file is saved.

    References
    ----------
    .. [#] A. Heimes and M. Vorländer, “Bidirectional surface scattering
           coefficients,” Acta Acust., vol. 9, p. 41, 2025,
           doi: 10.1051/aacus/2025026.



    """
    if not isinstance(source_directions, pf.Coordinates):
        raise TypeError(
            'source_directions must be a pf.Coordinates object')
    if not isinstance(receiver_directions, pf.Coordinates):
        raise TypeError(
            'receiver_directions must be a pf.Coordinates object')
    if (
            not isinstance(directional_scattering, pf.FrequencyData) or
            not directional_scattering.cshape == (
                source_directions.csize, receiver_directions.csize)):
        raise TypeError(
            'directional_scattering must be a pf.FrequencyData object with'
            f' cshape ({source_directions.csize, receiver_directions.csize})')
    if absorption_coefficient is None:
        absorption_coefficient = pf.FrequencyData(
            np.zeros_like(directional_scattering.frequencies),
            directional_scattering.frequencies)
    cos_receiver = np.cos(receiver_directions.colatitude)[
        np.newaxis, :, np.newaxis]
    receiver_weights = receiver_directions.weights
    receiver_weights *= 2 * np.pi / np.sum(receiver_weights)
    receiver_factor = receiver_weights[..., np.newaxis]
    brdf = directional_scattering.freq / receiver_factor / cos_receiver

    brdf *= (1 - absorption_coefficient.freq.flatten())
    if file_path is not None:
        sofa = _create_sofa(
            pf.FrequencyData(brdf, directional_scattering.frequencies),
            source_directions,
            receiver_directions,
            history='constructed brdf based on directional scattering',
        )

        sf.write_sofa(file_path, sofa)
    return pf.FrequencyData(brdf, absorption_coefficient.frequencies)


def _create_sofa(
    data,
    sources,
    receivers,
    history,
):
    """Write complex pressure to a SOFA object.

    Note that it will also write down the weights for the sources and
    the receivers.

    Parameters
    ----------
    data : numpy array
        The data as an array of shape (MRE)
    sources : pf.Coordinates
        source positions containing weights.
    receivers : pf.Coordinates
        receiver positions containing weights.
    history : string
        GLOBAL_History tag in the SOFA file.


    Returns
    -------
    sofa : sofar.Sofa object
        SOFA object with the data written to it

    """
    # create empty SOFA object
    convention = (
        'GeneralTF' if type(data) is pf.FrequencyData else 'GeneralFIR'
    )

    sofa = sf.Sofa(convention)

    # write meta data
    sofa.GLOBAL_ApplicationName = 'sparrowpy'
    sofa.GLOBAL_ApplicationVersion = sparrowpy.__version__
    sofa.GLOBAL_History = history

    # Source and receiver data
    sofa.EmitterPosition = sources.cartesian
    sofa.EmitterPosition_Units = 'meter'
    sofa.EmitterPosition_Type = 'cartesian'

    sources_sph = sources.spherical_elevation
    sources_sph = pf.rad2deg(sources_sph)
    sofa.SourcePosition = sources_sph
    sofa.SourcePosition_Units = 'degree, degree, metre'
    sofa.SourcePosition_Type = 'spherical'

    sofa.ReceiverPosition = receivers.cartesian
    sofa.ReceiverPosition_Units = 'meter'
    sofa.ReceiverPosition_Type = 'cartesian'

    if type(data) is pf.FrequencyData:
        sofa.N = data.frequencies

        # HRTF/HRIR data
        if data.cshape[0] != sources.csize:
            data.freq = np.swapaxes(data.freq, 0, 1)
        sofa.Data_Real = np.real(data.freq)
        sofa.Data_Imag = np.imag(data.freq)
    else:
        sofa.Data_IR = data.time
        sofa.Data_SamplingRate = data.sampling_rate
        sofa.Data_Delay = np.zeros((1, receivers.csize))

    sofa.add_variable('ReceiverWeights', receivers.weights, 'double', 'R')
    sofa.add_variable('SourceWeights', sources.weights, 'double', 'E')

    return sofa



"""Analytical solution for form factors for testing."""
import numpy as np

def perpendicular_patch_coincidentline(w, h, length):
    """
    Returns the exact form factor solution of two rectangular patches,
    The patches are orthogonal and share a side of length l
    w and h correspond to the width of each of the patches.
    """

    x = w/length
    y = h/length

    z = np.sqrt(x**2 + y**2)

    k = 1/(np.pi * y)

    a = x*np.arctan(1/x)
    b = y*np.arctan(1/y)
    c = -z*np.arctan(1/z)

    dk = 1/4
    da = (1+x**2)*(1+y**2)/(1+z**2)
    db = ( x**2 * (1+z**2) / (z**2 * (1 + x**2)) )**(x**2)
    dc = ( y**2 * (1+z**2) / (z**2 * (1 + y**2)) )**(y**2)

    d = dk*np.log(da*db*dc)

    return k * ( a + b + c + d )


def perpendicular_patch_coincidentpoint(aa,bb,cc,dd):
    """
    Returns the exact form factor solution of two rectangular patches,
    The patches are orthogonal and share a vertex.
    """

    a = aa/(bb+dd)
    b = cc/(bb+dd)
    c = np.sqrt(a**2+b**2)
    d = aa/bb
    e = cc/bb
    f = np.sqrt(d**2+e**2)
    x = aa/dd
    y = cc/dd
    z = np.sqrt(x**2+y**2)

    k = 1/(8*np.pi*bb*cc)

    out  = (bb+dd)**2 * np.log(((c**2 + a**2*b**2 + 1)/(c**2 +1)) * ((
        b**2 + b**2*c**2)/(c**2 +c**2*b**2))**(b**2) * ((
            a**2 + a**2*c**2)/(c**2 +c**2*a**2))**(a**2))

    out += bb**2 * np.log(((f**2 + 1)/(f**2 + d**2*e**2 + 1)) * ((
        f**2 + f**2*e**2)/(e**2 + e**2*f**2))**(e**2) * ((
            f**2 + f**2*d**2)/(d**2 + d**2*f**2))**(d**2))

    out += dd**2 * np.log(((z**2 + 1)/(z**2 + x**2*y**2 + 1)) * ((
        z**2 + z**2*y**2)/(y**2 + y**2*z**2))**(y**2) * ((
            z**2 + z**2*x**2)/(x**2 + x**2*z**2))**(x**2))

    out += 4 * (bb+dd)**2 * (
        a*np.arctan(1/a) + b*np.arctan(1/b) - c*np.arctan(1/c))

    out -= 4 * bb**2 * (
        d*np.arctan(1/d) + e*np.arctan(1/e) - f*np.arctan(1/f)  )

    out -= 4 * dd**2 * (
        x*np.arctan(1/x) + y*np.arctan(1/y) - z*np.arctan(1/z))

    return out * k


def perpendicular_patch_floating(aa,bb,cc,dd,ee):
    """
    Returns the exact form factor solution of two rectangular patches.
    The patches are orthogonal.
    """

    a = (dd+ee)/bb
    b = (aa+cc)/bb
    c = np.sqrt(aa**2 + bb**2)
    d = dd/bb
    y = cc/bb
    f = np.sqrt(d**2+b**2)
    j = np.sqrt(d**2+y**2)
    z = np.sqrt(a**2 + y**2)

    out = np.log(  ( (1+f**2)*(1+z**2) / ( (1+c**2)*(1+j**2) ) ) * ((
        z**2 + z**2 * c**2 )/(  c**2 + c**2 * z**2  ))**(a**2)  * ((
            f**2 + f**2 * c**2 )/(  c**2 + c**2 * f**2  ))**(b**2) * ((
                f**2 + f**2 * j**2 )/(  j**2 + j**2 * f**2  ))**(d**2) * ((
                    z**2 + z**2 * j**2 )/(  j**2 + j**2 * z**2  ))**(y**2)  )
    out += 4 * (
        f*np.arctan(1/f) + z*np.arctan(1/z) - c*np.arctan(
            1/c) - j*np.arctan(1/j))

    return out*bb/(4*aa*np.pi)

def parallel_patches(a,b,c):
    """
    Returns the exact form factor solution of parallel patches
    facing each other both patches have a length a and width b.
    They have a distance c from each other.
    """

    x = a/c

    y = b/c

    kk = 2 / ( np.pi * x * y )

    g = np.log( np.sqrt(( 1 + x**2 ) * ( 1 + y**2 ) / ( 1 + x**2 + y**2 )) )
    h = x * np.sqrt( 1 + y**2 ) * np.arctan( x / np.sqrt( 1 + y**2 ) )
    i = y * np.sqrt( 1 + x**2 ) * np.arctan( y / np.sqrt( 1 + x**2 ) )
    j = - x * np.arctan( x )
    k = - y * np.arctan( y )

    return kk * ( g + h + i + j + k)
